# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import cirq
import numpy as np

from mitiq import QPROGRAM
from mitiq.interface import convert_from_mitiq


def generate_random_clifford_t_circuit(
    num_qubits: int,
    num_oneq_cliffords: int,
    num_twoq_cliffords: int,
    num_t_gates: int,
    return_type: Optional[str] = None,
    seed: Optional[int] = None,
) -> QPROGRAM:
    r"""Generate a random quantum circuit with the given number of qubits,
    number of one-qubit Cliffords, number of two-qubit Cliffords and number
    of T gates.

    Args:
        num_qubits: The number of qubits in the generated circuit.
        num_oneq_cliffords: Number of one-qubit Cliffords to be used.
        num_twoq_cliffords: Number of two-qubit Cliffords to be used.
        num_t_gates: Number of T gates to be used.
        seed: Seed for generating random circuit.
        return_type: String which specifies the type of the returned
            circuits. See the keys of ``mitiq.SUPPORTED_PROGRAM_TYPES``
            for options. If ``None``, the returned circuits have type
            ``cirq.Circuit``.

    Returns:
        A quantum circuit acting on ``num_qubits`` qubits.
    """

    if num_qubits <= 0:
        raise ValueError(
            "Cannot prepare a circuit with {} qubits.", num_qubits
        )
    elif num_qubits == 1 and num_twoq_cliffords > 0:
        raise ValueError(
            "Need more than 2 qubits for two-qubit Clifford gates."
        )

    rnd_state = np.random.RandomState(seed)

    oneq_cliffords = [cirq.S, cirq.H]
    twoq_cliffords = [cirq.CNOT, cirq.CZ]

    oneq_idx_list = rnd_state.choice(len(oneq_cliffords), num_oneq_cliffords)
    twoq_idx_list = rnd_state.choice(len(twoq_cliffords), num_twoq_cliffords)

    oneq_list = [oneq_cliffords[i] for i in oneq_idx_list]
    twoq_list = [twoq_cliffords[i] for i in twoq_idx_list]

    t_list = [cirq.T for _ in range(num_t_gates)]

    all_gates = oneq_list + twoq_list + t_list
    rnd_state.shuffle(all_gates)

    qubits = cirq.LineQubit.range(num_qubits)
    circuit = cirq.Circuit()

    qubits_idx = list(range(num_qubits))

    for gate in all_gates:
        qubits_for_gate_idx = rnd_state.choice(
            qubits_idx, size=gate.num_qubits(), replace=False
        )
        qubits_for_gate = [qubits[i] for i in qubits_for_gate_idx]
        operation = gate.on(*qubits_for_gate)
        circuit.append(operation)

    return_type = "cirq" if not return_type else return_type
    return convert_from_mitiq(circuit, return_type)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions to convert between Mitiq's internal circuit representation and
Qibo's circuit representation.
"""

from typing import List

from cirq import Circuit
from numpy import pi
from qibo import gates
from qibo.gates.abstract import Gate, ParametrizedGate
from qibo.models.circuit import Circuit as QiboCircuit

from mitiq.interface.mitiq_qiskit import from_qasm as cirq_from_qasm
from mitiq.interface.mitiq_qiskit import to_qasm as cirq_to_qasm


def crx_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes CRX gate to Cirq known gates.

    Args:
        qibo_gate: CRX gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = gate.parameters[0]
    decomp_gate = [
        gates.RZ(q1, pi / 2),
        gates.RY(q1, theta / 2),
        gates.CNOT(q0, q1),
        gates.RY(q1, -theta / 2),
        gates.CNOT(q0, q1),
        gates.RZ(q1, -pi / 2),
    ]
    return decomp_gate


def cry_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes CRY gate to Cirq known gates.

    Args:
        gate: CRY gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = gate.parameters[0]
    decomp_gate = [
        gates.RY(q1, theta / 2),
        gates.CNOT(q0, q1),
        gates.RY(q1, -theta / 2),
        gates.CNOT(q0, q1),
    ]
    return decomp_gate


def crz_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes CRZ gate to Cirq known gates.

    Args:
        gate: CRZ gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = gate.parameters[0]
    decomp_gate = [
        gates.RZ(q1, theta / 2),
        gates.CNOT(q0, q1),
        gates.RZ(q1, -theta / 2),
        gates.CNOT(q0, q1),
    ]
    return decomp_gate


def cu1_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes CU1 gate to Cirq known gates.

    Args:
        gate: CU1 gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = gate.parameters[0]
    decomp_gate = [
        gates.U1(q0, theta / 2),
        gates.CNOT(q0, q1),
        gates.U1(q1, -theta / 2),
        gates.CNOT(q0, q1),
        gates.U1(q1, theta / 2),
    ]
    return decomp_gate


def cu3_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes CU3 gate to Cirq known gates.

    Args:
        gate: CU3 gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = gate.parameters[0]
    phi = gate.parameters[1]
    lam = gate.parameters[2]
    decomp_gate = [
        gates.U1(q0, lam / 2 + phi / 2),
        gates.U1(q1, lam / 2 - phi / 2),
        gates.CNOT(q0, q1),
        gates.U3(q1, -theta / 2, 0, -lam / 2 - phi / 2),
        gates.CNOT(q0, q1),
        gates.U3(q1, theta / 2, phi, 0),
    ]
    return decomp_gate


def csx_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes CSX gate to Cirq known gates.

    Args:
        gate: CSX gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = pi / 2
    decomp_gate = [
        gates.H(q1),
        gates.U1(q0, theta / 2),
        gates.CNOT(q0, q1),
        gates.U1(q1, -theta / 2),
        gates.CNOT(q0, q1),
        gates.U1(q1, theta / 2),
        gates.H(q1),
    ]
    return decomp_gate


def csxdg_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes CSXDG gate to Cirq known gates.

    Args:
        gate: CSXDG gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = -pi / 2
    decomp_gate = [
        gates.H(q1),
        gates.U1(q0, theta / 2),
        gates.CNOT(q0, q1),
        gates.U1(q1, -theta / 2),
        gates.CNOT(q0, q1),
        gates.U1(q1, theta / 2),
        gates.H(q1),
    ]
    return decomp_gate


def iswap_decomp(gate: Gate) -> List[Gate]:
    """Decomposes ISWAP gate to Cirq known gates.

    Args:
        gate: ISWAP gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    decomp_gate = [
        gates.S(q1),
        gates.S(q0),
        gates.H(q0),
        gates.CNOT(q0, q1),
        gates.CNOT(q1, q0),
        gates.H(q1),
    ]
    return decomp_gate


def fswap_decomp(gate: Gate) -> List[Gate]:
    """Decomposes FSWAP gate to Cirq known gates.

    Args:
        gate: FSWAP gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    decomp_gate = [
        gates.H(q1),
        gates.H(q0),
        gates.CNOT(q1, q0),
        gates.RZ(q0, pi / 2),
        gates.CNOT(q1, q0),
        gates.H(q1),
        gates.H(q0),
        gates.RX(q0, pi / 2),
        gates.RX(q1, pi / 2),
        gates.CNOT(q1, q0),
        gates.RZ(q0, pi / 2),
        gates.CNOT(q1, q0),
        gates.RX(q0, -pi / 2),
        gates.RX(q1, -pi / 2),
        gates.RZ(q0, pi / 2),
        gates.RZ(q1, pi / 2),
    ]
    return decomp_gate


def rxx_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes RXX gate to Cirq known gates.

    Args:
        gate: RXX gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = gate.parameters[0]
    decomp_gate = [
        gates.H(q0),
        gates.H(q1),
        gates.CNOT(q0, q1),
        gates.RZ(q1, theta),
        gates.CNOT(q0, q1),
        gates.H(q0),
        gates.H(q1),
    ]
    return decomp_gate


def ryy_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes RYY gate to Cirq known gates.

    Args:
        gate: RYY gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = gate.parameters[0]
    decomp_gate = [
        gates.RX(q0, pi / 2),
        gates.RX(q1, pi / 2),
        gates.CNOT(q0, q1),
        gates.RZ(q1, theta),
        gates.CNOT(q0, q1),
        gates.RX(q0, -pi / 2),
        gates.RX(q1, -pi / 2),
    ]
    return decomp_gate


def rzz_decomp(gate: ParametrizedGate) -> List[Gate]:
    """Decomposes RZZ gate to Cirq known gates.

    Args:
        gate: RZZ gate to decompose.

    Returns:
        List with gates that has the same effect as applying the original gate.
    """
    q0, q1 = gate.init_args
    theta = gate.parameters[0]
    decomp_gate = [
        gates.CNOT(q0, q1),
        gates.RZ(q1, theta),
        gates.CNOT(q0, q1),
    ]
    return decomp_gate


GATES_TO_DECOMPOSE = {
    "crx": crx_decomp,
    "cry": cry_decomp,
    "crz": crz_decomp,
    "cu1": cu1_decomp,
    "cu3": cu3_decomp,
    "csx": csx_decomp,
    "csxdg": csxdg_decomp,
    "iswap": iswap_decomp,
    "fswap": fswap_decomp,
    "rxx": rxx_decomp,
    "ryy": ryy_decomp,
    "rzz": rzz_decomp,
}


class UnsupportedQiboCircuitError(Exception):
    pass


def decompose_qibo_circuit(qibo_circuit: QiboCircuit) -> QiboCircuit:
    """Returns a QiboCircuit circuit equivalent to the input QiboCircuit
    with all unknown by cirq gates decomposed.

    Args:
        qibo_circuit: QiboCircuit to decompose.

    Returns:
        Decomposed QiboCircuit
    """
    decomp_circuit = QiboCircuit(qibo_circuit.nqubits)
    for gate in qibo_circuit.queue:
        decomposition_func = GATES_TO_DECOMPOSE.get(gate.name, lambda x: x)
        decomp_circuit.add(decomposition_func(gate))

    return decomp_circuit


def from_qibo(qibo_circuit: QiboCircuit) -> Circuit:
    """Returns a Cirq circuit equivalent to the input QiboCircuit.

    Args:
        qibo_circuit: QiboCircuit to convert to a Cirq circuit.

    Returns:
        Cirq circuit representation equivalent to the input QiboCircuit.
    """
    for measurement in qibo_circuit.measurements:
        reg_name = measurement.register_name
        if not reg_name.islower():
            raise UnsupportedQiboCircuitError(
                f"OpenQASM does not support capital letters in "
                f"register names but {reg_name} was used."
            )
    for gate in qibo_circuit.queue:
        if isinstance(gate, gates.M):
            continue
        if gate.is_controlled_by:
            raise UnsupportedQiboCircuitError(
                "OpenQASM does not support multi-controlled gates."
            )
        try:
            gate.qasm_label
        except NotImplementedError:
            raise UnsupportedQiboCircuitError(
                f"{gate.name} is not supported by OpenQASM."
            )

    # Decompose the circuit in known Cirq gates
    qasm = decompose_qibo_circuit(qibo_circuit).to_qasm()

    return cirq_from_qasm(qasm)


def to_qibo(circuit: Circuit) -> QiboCircuit:
    """Returns a QiboCircuit equivalent to the input Cirq circuit.

    Args:
        circuit: Cirq circuit to convert to a QiboCircuit.

    Returns:
        QiboCircuit object equivalent to the input Mitiq circuit.
    """
    qasm = cirq_to_qasm(circuit)
    # Remove problematic comment lines in the qasm code
    lines = [
        line.replace(", ", ",")
        for line in qasm.split("\n")
        if not line.strip().startswith("//")
    ]
    # Remove in line problematic comments
    clean_lines = []
    for line in lines:
        clean_line = line.split("//")[0].strip()
        if clean_line:
            clean_lines.append(clean_line)
    qasm = "\n".join(clean_lines)

    return QiboCircuit.from_qasm(qasm)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable

from mitiq import Bitstring, MeasurementResult


def post_select(
    result: MeasurementResult,
    selector: Callable[[Bitstring], bool],
    inverted: bool = False,
) -> MeasurementResult:
    """Returns only the bitstrings which satisfy the predicate in ``selector``.

    Args:
        result: List of bitstrings.
        selector: Predicate for which bitstrings to select. Examples:

            * ``selector = lambda bitstring: sum(bitstring) == k``
              - Select all bitstrings of Hamming weight ``k``.
            * ``selector = lambda bitstring: sum(bitstring) <= k``
              - Select all bitstrings of Hamming weight at most ``k``.
            * ``selector = lambda bitstring: bitstring[0] == 1``
              - Select all bitstrings such that the the first bit is 1.

        inverted: Invert the selector predicate so that bitstrings which obey
            ``selector(bitstring) == False`` are selected and returned.
    """
    return MeasurementResult(
        [bits for bits in result.result if selector(bits) != inverted]
    )



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Methods for scaling noise in circuits by layers and using multivariate extrapolation."""

from mitiq.lre.lre import execute_with_lre, mitigate_executor, lre_decorator, construct_circuits, combine_results

from mitiq.lre import multivariate_scaling

from mitiq.lre import inference



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions to convert between Mitiq's internal circuit representation and
Pennylane's circuit representation.
"""

from cirq import Circuit
from pennylane import from_qasm as pennylane_from_qasm
from pennylane.tape import QuantumTape
from pennylane.wires import Wires
from pennylane_qiskit.qiskit_device import QISKIT_OPERATION_MAP

from mitiq.interface.mitiq_qiskit import from_qasm as cirq_from_qasm
from mitiq.interface.mitiq_qiskit import to_qasm

SUPPORTED_PL = set(QISKIT_OPERATION_MAP.keys())
UNSUPPORTED = {"CRX", "CRY", "CRZ", "S", "T"}
SUPPORTED = SUPPORTED_PL - UNSUPPORTED


class UnsupportedQuantumTapeError(Exception):
    pass


def from_pennylane(tape: QuantumTape) -> Circuit:
    """Returns a Mitiq circuit equivalent to the input QuantumTape.

    Args:
        tape: Pennylane QuantumTape to convert to a Mitiq circuit.

    Returns:
        Mitiq circuit representation equivalent to the input QuantumTape.
    """
    try:
        wires = sorted(tape.wires)
    except TypeError:
        raise UnsupportedQuantumTapeError(
            f"The wires of the tape must be sortable, but could not sort "
            f"{tape.wires}."
        )

    for i in range(len(wires)):
        if wires[i] != i:
            raise UnsupportedQuantumTapeError(
                "The wire labels of the tape must contiguously pack 0 "
                "to n-1, for n wires."
            )

    if len(tape.measurements) > 0:
        raise UnsupportedQuantumTapeError(
            "Measurements are not supported on the input tape. "
            "They should be subsequently added by the executor."
        )

    tape = tape.expand(stop_at=lambda obj: obj.name in SUPPORTED)
    qasm = tape.to_openqasm(rotations=False, wires=wires, measure_all=False)

    return cirq_from_qasm(qasm)


def to_pennylane(circuit: Circuit) -> QuantumTape:
    """Returns a QuantumTape equivalent to the input Mitiq circuit.

    Args:
        circuit: Mitiq circuit to convert to a Pennylane QuantumTape.

    Returns:
        QuantumTape object equivalent to the input Mitiq circuit.
    """
    qfunc = pennylane_from_qasm(to_qasm(circuit))

    with QuantumTape() as tape:
        qfunc(wires=Wires(range(len(circuit.all_qubits()))))

    return tape



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from mitiq.interface.mitiq_cirq.cirq_utils import (
    sample_bitstrings,
    compute_density_matrix,
    execute_with_depolarizing_noise,
)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Readout Confusion Inversion."""

from functools import wraps
from types import MethodType
from typing import Callable, List, Sequence, Union, cast

import numpy as np
import numpy.typing as npt

from mitiq import QPROGRAM, MeasurementResult
from mitiq.executor.executor import Executor
from mitiq.observable.observable import Observable
from mitiq.rem.inverse_confusion_matrix import mitigate_measurements


def execute_with_rem(
    circuit: QPROGRAM,
    executor: Union[Executor, Callable[[QPROGRAM], MeasurementResult]],
    observable: Observable,
    *,
    inverse_confusion_matrix: npt.NDArray[np.float64],
) -> float:
    """Returns the readout error mitigated expectation value utilizing an
    inverse confusion matrix.

    Args:
        executor: A Mitiq executor that executes a circuit and returns the
            unmitigated ``MeasurementResult``.
        observable: Observable to compute the expectation value of (required).
        inverse_confusion_matrix: The inverse confusion matrix to apply to the
            probability vector estimated with noisy measurement results.

    Returns:
        The expectation value estimated with REM.
    """
    if not isinstance(executor, Executor):
        executor_obj = Executor(executor)
    else:
        executor_obj = executor

    executor_with_rem = mitigate_executor(
        executor_obj, inverse_confusion_matrix=inverse_confusion_matrix
    )
    # Since the input is an Executor the output is an Executor
    executor_with_rem = cast(Executor, executor_with_rem)

    return executor_with_rem.evaluate(circuit, observable)[0]


def mitigate_executor(
    executor: Union[Executor, Callable[[QPROGRAM], MeasurementResult]],
    *,
    inverse_confusion_matrix: npt.NDArray[np.float64],
) -> Union[Executor, Callable[[QPROGRAM], MeasurementResult]]:
    """Returns a modified version of the input 'executor' which is
    error-mitigated with readout confusion inversion (RCI).
    The type of the output executor will be equal to the type of
    the input executor: an :class:`.Executor` object or a Python callable.

    Args:
        executor: A Mitiq executor that executes a circuit and returns the
            unmitigated ``MeasurementResult``.
        inverse_confusion_matrix: The inverse confusion matrix to apply to the
            probability vector estimated with noisy measurement results.

    Returns:
        The error-mitigated version of the input executor.
    """
    # We always mitigate an Executor object but, to preserve the input type,
    # we eventually return a callable if the input executor is a callable.
    if not isinstance(executor, Executor):
        executor_obj = Executor(executor)
    else:
        executor_obj = executor

    def post_run(
        self: Executor,
        results: Sequence[MeasurementResult],
    ) -> Sequence[MeasurementResult]:
        return [
            mitigate_measurements(res, inverse_confusion_matrix)
            for res in results
        ]

    setattr(executor_obj, "_post_run", MethodType(post_run, executor_obj))

    if isinstance(executor, Executor):
        new_executor = executor_obj

    elif not executor_obj.can_batch:

        @wraps(executor)
        def new_executor(circuit: QPROGRAM) -> MeasurementResult:
            result = cast(MeasurementResult, executor_obj.run([circuit])[0])
            return result

    elif executor_obj.can_batch:

        @wraps(executor)
        def new_executor(
            circuits: List[QPROGRAM],
        ) -> Sequence[MeasurementResult]:
            results = executor_obj.run(circuits)
            return cast(Sequence[MeasurementResult], results)

    return new_executor


def rem_decorator(
    *,
    inverse_confusion_matrix: npt.NDArray[np.float64],
) -> Callable[
    [Callable[[QPROGRAM], MeasurementResult]],
    Callable[[QPROGRAM], MeasurementResult],
]:
    """Decorator which adds an error-mitigation layer based on readout
    confusion inversion (RCI) to an executor function, i.e., a function
    which executes a quantum circuit with an arbitrary backend and returns
    a ``MeasurementResult``.

    Args:
        inverse_confusion_matrix: The inverse confusion matrix to apply to the
            probability vector estimated with noisy measurement results
            (required).

    Returns:
        The error-mitigating decorator to be applied to an executor function.
    """
    # NOTE: most decorators check for whether the decorator has been used
    #   without parenthesis, but that is not possible with this decorator
    #   since arguments are required.

    def decorator(
        executor: Callable[[QPROGRAM], MeasurementResult],
    ) -> Callable[[QPROGRAM], MeasurementResult]:
        mitigated_executor = mitigate_executor(
            executor,
            inverse_confusion_matrix=inverse_confusion_matrix,
        )
        return cast(
            Callable[[QPROGRAM], MeasurementResult],
            mitigated_executor,
        )

    return decorator



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Built-in rules determining what DDD sequence should be applied in a given slack window."""
from mitiq.ddd.rules.rules import xx, xyxy, yy, general_rule, repeated_rule



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Methods for scaling noise in circuits by adding or modifying gates."""
from mitiq.zne.scaling.folding import (
    fold_all,
    fold_gates_at_random,
    fold_global,
)
from mitiq.zne.scaling.layer_scaling import (
    layer_folding,
    get_layer_folding,
)
from mitiq.zne.scaling.parameter import (
    scale_parameters,
    compute_parameter_variance,
)
from mitiq.zne.scaling.identity_insertion import insert_id_layers



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions for creating a linear complexity W-state benchmarking circuit
as defined in :cite:`Cruz_2019_Efficient`."""

from typing import Optional

import cirq
import numpy as np

from mitiq import QPROGRAM
from mitiq.interface import convert_from_mitiq


def generate_w_circuit(
    n_qubits: int,
    return_type: Optional[str] = None,
) -> QPROGRAM:
    """Returns a circuit to create a ``n_qubits`` qubit Werner-state with
    linear complexity as defined in :cite:`Cruz_2019_Efficient`.


    Args:
        n_qubits : The number of qubits in the circuit.
        return_type : Return type of the output circuit.
    Returns:
        A W-state circuit of linear complexity acting on ``n_qubits`` qubits.
    """
    if n_qubits <= 0:
        raise ValueError("{} is invalid for the number of qubits. ", n_qubits)

    qubits = cirq.LineQubit.range(n_qubits)
    circuit = cirq.Circuit()

    for i, j in zip(range(0, n_qubits), range(1, n_qubits)):
        N = n_qubits - i
        angle = 2 * np.arccos(np.sqrt(1 / N))
        circuit.append(
            cirq.Ry(rads=angle).controlled().on(qubits[i], qubits[j])
        )
        circuit.append(cirq.CNOT(qubits[j], qubits[i]))

    return_type = "cirq" if not return_type else return_type

    return convert_from_mitiq(circuit, return_type)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions for mapping circuits to (near) Clifford circuits."""

from typing import Any, List, Optional, Sequence, Union, cast

import cirq
import numpy as np
from cirq.circuits import Circuit

from mitiq.cdr.clifford_utils import (
    angle_to_proximity,
    closest_clifford,
    probabilistic_angle_to_clifford,
    random_clifford,
)
from mitiq.interface import atomic_one_to_many_converter


@atomic_one_to_many_converter
def generate_training_circuits(
    circuit: Circuit,
    num_training_circuits: int,
    fraction_non_clifford: float,
    method_select: str = "uniform",
    method_replace: str = "closest",
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    **kwargs: Any,
) -> List[Circuit]:
    r"""Returns a list of (near) Clifford circuits obtained by replacing (some)
    non-Clifford gates in the input circuit by Clifford gates.
    The way in which non-Clifford gates are selected to be replaced is
    determined by ``method_select`` and ``method_replace``.
    In the Clifford Data Regression (CDR) method
    :cite:`Czarnik_2021_Quantum`, data generated from these circuits is used
    as a training set to learn the effect of noise.

    Args:
        circuit: A circuit of interest assumed to be compiled into the gate
            set {Rz, sqrt(X), CNOT}, or such that all the non-Clifford gates
            are contained in the Rz rotations.
        num_training_circuits: Number of circuits in the returned training set.
        fraction_non_clifford: The (approximate) fraction of non-Clifford
            gates in each returned circuit.
        method_select: Method by which non-Clifford gates are selected to be
            replaced by Clifford gates. Options are 'uniform' or 'gaussian'.
        method_replace: Method by which selected non-Clifford gates are
            replaced by Clifford gates. Options are 'uniform', 'gaussian' or
            'closest'.
        random_state: Seed for sampling.
        kwargs: Available keyword arguments are:

            - sigma_select (float): Width of the Gaussian distribution used for
              ``method_select='gaussian'``.

            - sigma_replace (float): Width of the Gaussian distribution used
              for ``method_replace='gaussian'``.
    """
    if random_state is None or isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)

    # Find the non-Clifford operations in the circuit.
    operations = list(circuit.all_operations())
    non_clifford_indices_and_ops = [
        (i, op)
        for i, op in enumerate(operations)
        if not cirq.has_stabilizer_effect(op)
    ]

    if len(non_clifford_indices_and_ops) == 0:
        return [circuit] * num_training_circuits

    non_clifford_indices = [i for i, _ in non_clifford_indices_and_ops]
    non_clifford_ops = [op for _, op in non_clifford_indices_and_ops]

    # Replace (some of) the non-Clifford operations.
    near_clifford_circuits = []
    for _ in range(num_training_circuits):
        new_ops = _map_to_near_clifford(
            non_clifford_ops,
            fraction_non_clifford,
            method_select,
            method_replace,
            random_state,
            **kwargs,
        )
        for non_clifford_index, new_op in zip(non_clifford_indices, new_ops):
            operations[non_clifford_index] = new_op

        near_clifford_circuits.append(Circuit(operations))

    return near_clifford_circuits


def _map_to_near_clifford(
    non_clifford_ops: Sequence[cirq.ops.Operation],
    fraction_non_clifford: float,
    method_select: str = "uniform",
    method_replace: str = "closest",
    random_state: Optional[np.random.RandomState] = None,
    **kwargs: Any,
) -> Sequence[cirq.ops.Operation]:
    """Returns the list of non-Clifford operations with some of these replaced
    by Clifford operations.

    Args:
        non_clifford_ops: A sequence of non-Clifford operations.
        fraction_non_clifford: The (approximate) fraction of non-Clifford
            operations in the returned list.
        method_select: The way in which the non-Clifford gates are selected to
            be replaced by Clifford gates. Options are 'uniform' or 'gaussian'.
        method_replace: The way in which selected non-Clifford gates are
            replaced by Clifford gates. Options are 'uniform', 'gaussian' or
            'closest'.
        random_state: Seed for sampling.
        kwargs: Additional options for selection / replacement methods.

            - sigma_select (float): Width of the Gaussian distribution used for
            ``method_select='gaussian'``.

            - sigma_replace (float): Width of the Gaussian distribution used
            for ``method_replace='gaussian'``.
    """
    sigma_select: float = kwargs.get("sigma_select", 0.5)
    sigma_replace: float = kwargs.get("sigma_replace", 0.5)

    # Select (indices of) operations to replace.
    indices_of_selected_ops = _select(
        non_clifford_ops,
        fraction_non_clifford,
        method_select,
        sigma_select,
        random_state,
    )

    # Replace selected operations.
    clifford_ops = _replace(
        [non_clifford_ops[i] for i in indices_of_selected_ops],
        method_replace,
        sigma_replace,
        random_state,
    )

    # Return sequence of (near) Clifford operations.
    return [
        clifford_ops.pop(0) if i in indices_of_selected_ops else op
        for (i, op) in enumerate(non_clifford_ops)
    ]


def _select(
    non_clifford_ops: Sequence[cirq.ops.Operation],
    fraction_non_clifford: float,
    method: str = "uniform",
    sigma: Optional[float] = 1.0,
    random_state: Optional[np.random.RandomState] = None,
) -> List[int]:
    """Returns indices of non-Clifford operations selected (to be replaced)
    according to some method.

    Args:
        non_clifford_ops: Sequence of non-Clifford operations.
        fraction_non_clifford: fraction of non-Clifford gates to change.
        method: {'uniform', 'gaussian'} method to use to select Clifford gates
                to replace.
        sigma: width of probability distribution used in selection
                      of non-Clifford gates to replace, only has effect if
                      method_select = 'gaussian'.
        random_state: Random state for sampling.
    """
    if random_state is None:
        random_state = np.random  # type: ignore

    num_non_cliff = len(non_clifford_ops)
    num_to_replace = int(round(fraction_non_clifford * num_non_cliff))

    # Get the distribution for how to select operations.
    if method == "uniform":
        distribution = 1.0 / num_non_cliff * np.ones(shape=(num_non_cliff,))
    elif method == "gaussian":
        non_clifford_angles = np.array(
            [
                op.gate.exponent * np.pi  # type: ignore
                for op in non_clifford_ops
            ]
        )
        probabilities = angle_to_proximity(non_clifford_angles, sigma)
        distribution = probabilities / sum(probabilities)
    else:
        raise ValueError(
            f"Arg `method_select` must be 'uniform' or 'gaussian' but was "
            f"{method}."
        )

    # Select (indices of) non-Clifford operations to replace.
    selected_indices = cast(np.random.RandomState, random_state).choice(
        range(num_non_cliff),
        num_non_cliff - num_to_replace,
        replace=False,
        p=distribution,
    )
    return [int(i) for i in sorted(selected_indices)]


def _replace(
    non_clifford_ops: Sequence[cirq.ops.Operation],
    method: str = "uniform",
    sigma: float = 1.0,
    random_state: Optional[np.random.RandomState] = None,
) -> List[cirq.ops.Operation]:
    """Function that takes the non-Clifford angles and replacement and
    selection specifications, returning the projected angles according to a
    specific method.

    Args:
        non_clifford_ops: array of non-Clifford angles.
        method: {'uniform', 'gaussian', 'closest'} method to use
                        to replace selected non-Clifford gates.
        sigma: width of probability distribution used in replacement
                       of selected non-Clifford gates, only has effect if
                       method_replace = 'gaussian'.
        random_state: Seed for sampling.

    Returns:
        rz_non_clifford_replaced: the selected non-Clifford gates replaced by a
                               Clifford according to some method.

    Raises:
        Exception: If argument 'method_replace' is not either 'closest',
        'uniform' or 'gaussian'.
    """
    if random_state is None:
        random_state = np.random  # type: ignore

    # TODO: Update these functions to act on operations instead of angles.
    non_clifford_angles = np.array(
        [op.gate.exponent * np.pi for op in non_clifford_ops]  # type: ignore
    )
    if method == "closest":
        clifford_angles = closest_clifford(non_clifford_angles)

    elif method == "uniform":
        clifford_angles = random_clifford(
            len(non_clifford_angles), cast(np.random.RandomState, random_state)
        )

    elif method == "gaussian":
        clifford_angles = probabilistic_angle_to_clifford(
            non_clifford_angles, sigma, random_state
        )

    else:
        raise ValueError(
            f"Arg `method_replace` must be 'closest', 'uniform', or 'gaussian'"
            f" but was {method}."
        )

    # TODO: Write function to replace the angles in a list of operations?
    return [
        cirq.ops.rz(a).on(*q)
        for (a, q) in zip(
            clifford_angles,
            [op.qubits for op in non_clifford_ops],
        )
    ]



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions to convert between Mitiq's internal circuit representation and
pyQuil's circuit representation (Quil programs).
"""

from cirq import Circuit, LineQubit
from cirq_rigetti import circuit_from_quil
from cirq_rigetti.quil_output import QuilOutput
from pyquil import Program

QuilType = str


def to_quil(circuit: Circuit) -> QuilType:
    """Returns a Quil string representing the input Mitiq circuit.

    Args:
        circuit: Mitiq circuit to convert to a Quil string.

    Returns:
        QuilType: Quil string equivalent to the input Mitiq circuit.
    """
    max_qubit = max(circuit.all_qubits())
    # if we are using LineQubits, keep the qubit labeling the same
    if isinstance(max_qubit, LineQubit):
        qubit_range = max_qubit.x + 1
        return str(
            QuilOutput(circuit.all_operations(), LineQubit.range(qubit_range))
        )
    # otherwise, use the default ordering (starting from zero)
    return str(
        QuilOutput(circuit.all_operations(), sorted(circuit.all_qubits()))
    )


def to_pyquil(circuit: Circuit) -> Program:
    """Returns a pyQuil Program equivalent to the input Mitiq circuit.

    Args:
        circuit: Mitiq circuit to convert to a pyQuil Program.

    Returns:
        pyquil.Program object equivalent to the input Mitiq circuit.
    """
    return Program(to_quil(circuit))


def from_pyquil(program: Program) -> Circuit:
    """Returns a Mitiq circuit equivalent to the input pyQuil Program.

    Args:
        program: PyQuil Program to convert to a Mitiq circuit.

    Returns:
        Mitiq circuit representation equivalent to the input pyQuil Program.
    """
    return from_quil(program.out())


def from_quil(quil: QuilType) -> Circuit:
    """Returns a Mitiq circuit equivalent to the input Quil string.

    Args:
        quil: Quil string to convert to a Mitiq circuit.

    Returns:
        Mitiq circuit representation equivalent to the input Quil string.
    """
    return circuit_from_quil(quil)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from mitiq.pt.pt import (
    generate_pauli_twirl_variants,
    twirl_CNOT_gates,
    twirl_CZ_gates,
)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.
"""Cirq utility functions."""

from typing import Callable, Tuple

import cirq
import numpy as np
import numpy.typing as npt

from mitiq import MeasurementResult


# Executors.
def sample_bitstrings(
    circuit: cirq.Circuit,
    noise_model_function: Callable[
        ..., cirq.NOISE_MODEL_LIKE
    ] = cirq.amplitude_damp,
    noise_level: Tuple[float] = (0.01,),
    sampler: cirq.Sampler = cirq.DensityMatrixSimulator(),
    shots: int = 8192,
) -> MeasurementResult:
    """Adds noise to the input circuit. The noise is added based on a
    particular noise model and some value for the error rate.

    Args:
        circuit: The input Cirq circuit with measurements applied.
        noise_model: Input Cirq noise model. Default is amplitude damping.
        noise_level: Noise rate as a tuple of floats.
        sampler: Cirq simulator from which the result will be sampled from.
        shots: Number of measurements.

    Returns:
        Sampled outcome from a measurement.
    """
    if sum(noise_level) > 0:
        circuit = circuit.with_noise(noise_model_function(*noise_level))

    result = sampler.run(circuit, repetitions=shots)
    return MeasurementResult(
        result=np.column_stack(list(result.measurements.values())).tolist(),
        qubit_indices=tuple(
            # q[2:-1] is necessary to convert "q(number)" into "number"
            int(q[2:-1])
            for k in result.measurements.keys()
            for q in k.split(",")
        ),
    )


def compute_density_matrix(
    circuit: cirq.Circuit,
    noise_model_function: Callable[
        ..., cirq.NOISE_MODEL_LIKE
    ] = cirq.amplitude_damp,
    noise_level: Tuple[float] = (0.01,),
) -> npt.NDArray[np.complex64]:
    """Returns the density matrix of the quantum state after the
    (noisy) execution of the input circuit.

    Args:
        circuit: The input Cirq circuit.
        noise_model: Input Cirq noise model. Default is amplitude damping.
        noise_level: Noise rate as a tuple of floats.

    Returns:
        The final density matrix as a NumPy array.
    """
    if sum(noise_level) > 0:
        circuit = circuit.with_noise(noise_model_function(*noise_level))

    return cirq.DensityMatrixSimulator().simulate(circuit).final_density_matrix


def execute_with_depolarizing_noise(
    circuit: cirq.Circuit, obs: npt.NDArray[np.complex64], noise: float
) -> float:
    """Simulates a circuit with depolarizing noise
    and returns the expectation value of the input observable.
    The expectation value is deterministically computed from
    the final density matrix and, therefore, shot noise is absent.

    Args:
        circuit: The input Cirq circuit.
        obs: The observable to measure as a NumPy array.
        noise: The depolarizing noise as a float, i.e. 0.001 is 0.1% noise.

    Returns:
        The expectation value of obs as a float.
    """
    circuit = circuit.with_noise(cirq.depolarize(p=noise))
    simulator = cirq.DensityMatrixSimulator()
    rho = simulator.simulate(circuit).final_density_matrix
    expectation = np.real(np.trace(rho @ obs))
    return expectation



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.
from typing import List, Optional, Union, cast
from warnings import warn

import cirq_ionq.ionq_native_gates as cirq_ionq_ops
import numpy as np
import numpy.typing as npt
from braket.circuits import Circuit as BKCircuit
from braket.circuits import Instruction
from braket.circuits import gates as braket_gates
from cirq import Circuit, LineQubit, protocols
from cirq import ops as cirq_ops
from cirq.linalg.decompositions import (
    deconstruct_single_qubit_matrix_into_angles,
    kak_decomposition,
)


def _raise_braket_to_cirq_error(instr: Instruction) -> None:
    raise ValueError(
        f"Unable to convert the instruction {instr} to Cirq. If you think "
        "this is a bug, you can open an issue on the Mitiq GitHub at "
        "https://github.com/unitaryfoundation/mitiq."
    )


def _raise_cirq_to_braket_error(op: cirq_ops.Operation) -> None:
    raise ValueError(
        f"Unable to convert {op} to Braket. If you think this is a bug, "
        "you can open an issue on the Mitiq GitHub at"
        " https://github.com/unitaryfoundation/mitiq."
    )


def from_braket(circuit: BKCircuit) -> Circuit:
    """Returns a Cirq circuit equivalent to the input Braket circuit.

    Note: The returned Cirq circuit acts on cirq.LineQubit's with indices equal
    to the qubit indices of the Braket circuit.

    Args:
        circuit: Braket circuit to convert to a Cirq circuit.
    """
    return Circuit(
        _translate_braket_instruction_to_cirq_operation(instr)
        for instr in circuit.instructions
    )


def to_braket(circuit: Circuit) -> BKCircuit:
    """Returns a Braket circuit equivalent to the input Cirq circuit.

    Args:
        circuit: Cirq circuit to convert to a Braket circuit.
    """
    return BKCircuit(
        _translate_cirq_operation_to_braket_instruction(op)
        for op in circuit.all_operations()
    )


def _translate_braket_instruction_to_cirq_operation(
    instr: Instruction,
) -> List[cirq_ops.Operation]:
    """Converts the braket instruction to an equivalent Cirq operation or list
    of Cirq operations.

    Args:
        instr: Braket instruction to convert.

    Raises:
        ValueError: If the instruction cannot be converted to Cirq.
    """
    nqubits = len(instr.target)

    if nqubits == 1:
        return _translate_one_qubit_braket_instruction_to_cirq_operation(instr)

    elif nqubits == 2:
        return _translate_two_qubit_braket_instruction_to_cirq_operation(instr)

    elif nqubits == 3:
        qubits = [LineQubit(int(qubit)) for qubit in instr.target]

        if isinstance(instr.operator, braket_gates.CCNot):
            return [cirq_ops.TOFFOLI.on(*qubits)]
        elif isinstance(instr.operator, braket_gates.CSwap):
            return [cirq_ops.FREDKIN.on(*qubits)]
        else:
            _raise_braket_to_cirq_error(instr)

    # Unknown instructions.
    else:
        _raise_braket_to_cirq_error(instr)

    return None  # type: ignore[return-value]  # pragma: no cover


def _translate_cirq_operation_to_braket_instruction(
    op: cirq_ops.Operation,
) -> List[Instruction]:
    """Converts the Cirq operation to an equivalent Braket instruction or list
    of instructions.

    Args:
        op: Cirq operation to convert.

    Raises:
        ValueError: If the operation cannot be converted to Braket.
    """
    # Measurement gates do not exist in Braket.
    if isinstance(op.gate, cirq_ops.MeasurementGate):
        warn("Measurement gate removed when converting from Cirq to Braket.")
        return []

    nqubits = protocols.num_qubits(op)

    if nqubits == 1:
        return _translate_one_qubit_cirq_operation_to_braket_instruction(op)

    elif nqubits == 2:
        return _translate_two_qubit_cirq_operation_to_braket_instruction(op)

    elif nqubits == 3:
        qubits = [cast(LineQubit, q).x for q in op.qubits]

        if op == cirq_ops.TOFFOLI.on(*op.qubits):
            return [Instruction(braket_gates.CCNot(), qubits)]
        elif op == cirq_ops.FREDKIN.on(*op.qubits):
            return [Instruction(braket_gates.CSwap(), qubits)]
        else:
            _raise_cirq_to_braket_error(op)
    # Unsupported gates.
    else:
        _raise_cirq_to_braket_error(op)

    return None  # type: ignore[return-value]  # pragma: no cover


def _translate_one_qubit_braket_instruction_to_cirq_operation(
    instr: Instruction,
) -> List[cirq_ops.Operation]:
    """Converts the one-qubit braket instruction to Cirq.

    Args:
        instr: One-qubit Braket instruction to convert.

    Raises:
        ValueError: If the instruction cannot be converted to Cirq.
    """
    qubits = [LineQubit(int(qubit)) for qubit in instr.target]
    gate = instr.operator

    # One-qubit non-parameterized gates.
    if isinstance(gate, braket_gates.I):
        return [cirq_ops.I.on(*qubits)]
    elif isinstance(gate, braket_gates.X):
        return [cirq_ops.X.on(*qubits)]
    elif isinstance(gate, braket_gates.Y):
        return [cirq_ops.Y.on(*qubits)]
    elif isinstance(gate, braket_gates.Z):
        return [cirq_ops.Z.on(*qubits)]
    elif isinstance(gate, braket_gates.H):
        return [cirq_ops.H.on(*qubits)]
    elif isinstance(gate, braket_gates.S):
        return [cirq_ops.S.on(*qubits)]
    elif isinstance(gate, braket_gates.Si):
        return [protocols.inverse(cirq_ops.S.on(*qubits))]
    elif isinstance(gate, braket_gates.T):
        return [cirq_ops.T.on(*qubits)]
    elif isinstance(gate, braket_gates.Ti):
        return [protocols.inverse(cirq_ops.T.on(*qubits))]
    elif isinstance(gate, braket_gates.V):
        return [cirq_ops.X.on(*qubits) ** 0.5]
    elif isinstance(gate, braket_gates.Vi):
        return [cirq_ops.X.on(*qubits) ** -0.5]

    # One-qubit parameterized gates.
    elif isinstance(gate, braket_gates.Rx):
        return [cirq_ops.rx(gate.angle).on(*qubits)]
    elif isinstance(gate, braket_gates.Ry):
        return [cirq_ops.ry(gate.angle).on(*qubits)]
    elif isinstance(gate, braket_gates.Rz):
        return [cirq_ops.rz(gate.angle).on(*qubits)]
    elif isinstance(gate, braket_gates.PhaseShift):
        return [cirq_ops.Z.on(*qubits) ** (gate.angle / np.pi)]
    elif isinstance(gate, braket_gates.GPi):
        return [
            cirq_ionq_ops.GPIGate(phi=gate.angle / (2 * np.pi)).on(*qubits)
        ]
    elif isinstance(gate, braket_gates.GPi2):
        return [
            cirq_ionq_ops.GPI2Gate(phi=gate.angle / (2 * np.pi)).on(*qubits)
        ]

    else:
        _raise_braket_to_cirq_error(instr)

    return None  # type: ignore[return-value]  # pragma: no cover


def _translate_two_qubit_braket_instruction_to_cirq_operation(
    instr: Instruction,
) -> List[cirq_ops.Operation]:
    """Converts the two-qubit braket instruction to Cirq.

    Args:
        instr: Two-qubit Braket instruction to convert.

    Raises:
        ValueError: If the instruction cannot be converted to Cirq.
    """
    qubits = [LineQubit(int(qubit)) for qubit in instr.target]
    gate = instr.operator

    # Two-qubit non-parameterized gates.
    if isinstance(gate, braket_gates.CNot):
        return [cirq_ops.CNOT.on(*qubits)]

    elif isinstance(gate, braket_gates.Swap):
        return [cirq_ops.SWAP.on(*qubits)]
    elif isinstance(gate, braket_gates.ISwap):
        return [cirq_ops.ISWAP.on(*qubits)]
    elif isinstance(gate, braket_gates.CZ):
        return [cirq_ops.CZ.on(*qubits)]
    elif isinstance(gate, braket_gates.CY):
        return [
            protocols.inverse(cirq_ops.S.on(qubits[1])),
            cirq_ops.CNOT.on(*qubits),
            cirq_ops.S.on(qubits[1]),
        ]

    # Two-qubit parameterized gates.
    elif isinstance(gate, braket_gates.CPhaseShift):
        return [cirq_ops.CZ.on(*qubits) ** (gate.angle / np.pi)]
    elif isinstance(gate, braket_gates.CPhaseShift00):
        return [
            cirq_ops.XX(*qubits),
            cirq_ops.CZ.on(*qubits) ** (gate.angle / np.pi),
            cirq_ops.XX(*qubits),
        ]
    elif isinstance(gate, braket_gates.CPhaseShift01):
        return [
            cirq_ops.X(qubits[0]),
            cirq_ops.CZ.on(*qubits) ** (gate.angle / np.pi),
            cirq_ops.X(qubits[0]),
        ]
    elif isinstance(gate, braket_gates.CPhaseShift10):
        return [
            cirq_ops.X(qubits[1]),
            cirq_ops.CZ.on(*qubits) ** (gate.angle / np.pi),
            cirq_ops.X(qubits[1]),
        ]
    elif isinstance(gate, braket_gates.PSwap):
        return [
            cirq_ops.SWAP.on(*qubits),
            cirq_ops.CNOT.on(*qubits),
            cirq_ops.Z.on(qubits[1]) ** (gate.angle / np.pi),
            cirq_ops.CNOT.on(*qubits),
        ]
    elif isinstance(gate, braket_gates.XX):
        return [
            cirq_ops.XXPowGate(
                exponent=gate.angle / np.pi, global_shift=-0.5
            ).on(*qubits)
        ]
    elif isinstance(gate, braket_gates.YY):
        return [
            cirq_ops.YYPowGate(
                exponent=gate.angle / np.pi, global_shift=-0.5
            ).on(*qubits)
        ]
    elif isinstance(gate, braket_gates.ZZ):
        return [
            cirq_ops.ZZPowGate(
                exponent=gate.angle / np.pi, global_shift=-0.5
            ).on(*qubits)
        ]
    elif isinstance(gate, braket_gates.XY):
        return [cirq_ops.ISwapPowGate(exponent=gate.angle / np.pi).on(*qubits)]

    # Two-qubit two-parameters parameterized gates.
    elif isinstance(gate, braket_gates.MS):
        return [
            cirq_ionq_ops.MSGate(
                phi0=gate.angle_1 / (2 * np.pi),
                phi1=gate.angle_2 / (2 * np.pi),
            ).on(*qubits)
        ]

    else:
        _raise_braket_to_cirq_error(instr)

    return None  # type: ignore[return-value]  # pragma: no cover


def _translate_one_qubit_cirq_operation_to_braket_instruction(
    op: Union[npt.NDArray[np.complex64], cirq_ops.Operation],
    target: Optional[int] = None,
) -> List[Instruction]:
    """Translates a one-qubit Cirq operation to a (sequence of) Braket
    instruction(s) according to the following rules:

    1. Attempts to find a "standard translation" from Cirq to Braket.
        - e.g., checks if `op` is Pauli-X and, if so, returns the Braket X.

    2. If (1) is not successful, decomposes the unitary of `op` to
    Rz(theta) Ry(phi) Rz(lambda) and returns the series of rotations as Braket
    instructions.

    Args:
        op: One-qubit Cirq operation to translate.
        target: Qubit index for the op to act on. Must be specified and if only
            if `op` is given as a numpy array.
    """
    # Translate qubit index.
    if not isinstance(op, np.ndarray):
        target = cast(LineQubit, op.qubits[0]).x

    if target is None:
        raise ValueError(
            "Arg `target` must be specified when `op` is a matrix."
        )

    # Check common single-qubit gates.
    if isinstance(op, cirq_ops.Operation):
        if isinstance(op.gate, cirq_ops.XPowGate):
            exponent = cast(float, op.gate.exponent)
            if np.isclose(exponent, 1.0) or np.isclose(exponent, -1.0):
                return [Instruction(braket_gates.X(), target)]
            elif np.isclose(exponent, 0.5):
                return [Instruction(braket_gates.V(), target)]
            elif np.isclose(exponent, -0.5):
                return [Instruction(braket_gates.Vi(), target)]

            return [Instruction(braket_gates.Rx(exponent * np.pi), target)]

        elif isinstance(op.gate, cirq_ops.YPowGate):
            exponent = cast(float, op.gate.exponent)
            if np.isclose(exponent, 1.0) or np.isclose(exponent, -1.0):
                return [Instruction(braket_gates.Y(), target)]

            return [Instruction(braket_gates.Ry(exponent * np.pi), target)]

        elif isinstance(op.gate, cirq_ops.ZPowGate):
            exponent = cast(float, op.gate.exponent)
            if np.isclose(exponent, 1.0) or np.isclose(exponent, -1.0):
                return [Instruction(braket_gates.Z(), target)]
            elif np.isclose(exponent, 0.5):
                return [Instruction(braket_gates.S(), target)]
            elif np.isclose(exponent, -0.5):
                return [Instruction(braket_gates.Si(), target)]
            elif np.isclose(exponent, 0.25):
                return [Instruction(braket_gates.T(), target)]
            elif np.isclose(exponent, -0.25):
                return [Instruction(braket_gates.Ti(), target)]

            return [Instruction(braket_gates.Rz(exponent * np.pi), target)]

        elif isinstance(op.gate, cirq_ops.HPowGate) and np.isclose(
            abs(cast(float, op.gate.exponent)), 1.0
        ):
            return [Instruction(braket_gates.H(), target)]

        elif isinstance(op.gate, cirq_ops.IdentityGate):
            return [Instruction(braket_gates.I(), target)]

        # IonQ native gates
        elif isinstance(op.gate, cirq_ionq_ops.GPIGate):
            angle = op.gate.phi * 2 * np.pi
            return [Instruction(braket_gates.GPi(angle), target)]
        elif isinstance(op.gate, cirq_ionq_ops.GPI2Gate):
            angle = op.gate.phi * 2 * np.pi
            return [Instruction(braket_gates.GPi2(angle), target)]

    # Arbitrary single-qubit unitary decomposition.
    # TODO: This does not account for global phase.
    if isinstance(op, cirq_ops.Operation):
        unitary_matrix = protocols.unitary(op)
    else:
        unitary_matrix = op

    a, b, c = deconstruct_single_qubit_matrix_into_angles(unitary_matrix)
    return [
        Instruction(braket_gates.Rz(a), target),
        Instruction(braket_gates.Ry(b), target),
        Instruction(braket_gates.Rz(c), target),
    ]


def _translate_two_qubit_cirq_operation_to_braket_instruction(
    op: cirq_ops.Operation,
) -> List[Instruction]:
    """Translates a two-qubit Cirq operation to a (sequence of) Braket
    instruction(s) according to the following rules:

    1. Attempts to find a "standard translation" from Cirq to Braket.
        - e.g., checks if `op` is a CNOT and, if so, returns the Braket CNOT.

    2. If (1) is not successful, performs a KAK decomposition of the unitary of
    `op` to obtain a circuit

        ──A1──X^0.5───@───X^a───X──────────────────@───B1───
                      │         │                  │
        ──A2──────────X───Y^b───@───X^-0.5───Z^c───X───B2────

    where A1, A2, B1, and B2 are arbitrary single-qubit unitaries and a, b, c
    are floats.

    Args:
        op: Two-qubit Cirq operation to translate.
    """
    # Translate qubit indices.
    q1, q2 = [cast(LineQubit, qubit).x for qubit in op.qubits]

    # Check common two-qubit gates.
    if isinstance(op.gate, cirq_ops.CNotPowGate) and np.isclose(
        abs(cast(float, op.gate.exponent)), 1.0
    ):
        return [Instruction(braket_gates.CNot(), [q1, q2])]
    elif isinstance(op.gate, cirq_ops.CZPowGate) and np.isclose(
        abs(cast(float, op.gate.exponent)), 1.0
    ):
        return [Instruction(braket_gates.CZ(), [q1, q2])]
    elif isinstance(op.gate, cirq_ops.SwapPowGate) and np.isclose(
        cast(float, op.gate.exponent), 1.0
    ):
        return [Instruction(braket_gates.Swap(), [q1, q2])]
    elif isinstance(op.gate, cirq_ops.ISwapPowGate) and np.isclose(
        cast(float, op.gate.exponent), 1.0
    ):
        return [Instruction(braket_gates.ISwap(), [q1, q2])]
    elif isinstance(op.gate, cirq_ops.XXPowGate):
        return [
            Instruction(
                braket_gates.XX(cast(float, op.gate.exponent) * np.pi),
                [q1, q2],
            )
        ]
    elif isinstance(op.gate, cirq_ops.YYPowGate):
        return [
            Instruction(
                braket_gates.YY(cast(float, op.gate.exponent) * np.pi),
                [q1, q2],
            )
        ]
    elif isinstance(op.gate, cirq_ops.ZZPowGate):
        return [
            Instruction(
                braket_gates.ZZ(cast(float, op.gate.exponent) * np.pi),
                [q1, q2],
            )
        ]

    # IonQ native gates
    elif isinstance(op.gate, cirq_ionq_ops.MSGate):
        a0, a1 = op.gate.phi0, op.gate.phi1
        return [
            Instruction(
                braket_gates.MS(a0 * 2 * np.pi, a1 * 2 * np.pi),
                [q1, q2],
            )
        ]

    # Arbitrary two-qubit unitary decomposition.
    kak = kak_decomposition(protocols.unitary(op))
    A1, A2 = kak.single_qubit_operations_before

    x, y, z = kak.interaction_coefficients
    a = x * -2 / np.pi + 0.5
    b = y * -2 / np.pi + 0.5
    c = z * -2 / np.pi + 0.5

    B1, B2 = kak.single_qubit_operations_after

    return [
        *_translate_one_qubit_cirq_operation_to_braket_instruction(A1, q1),
        *_translate_one_qubit_cirq_operation_to_braket_instruction(A2, q2),
        Instruction(braket_gates.Rx(0.5 * np.pi), q1),
        Instruction(braket_gates.CNot(), [q1, q2]),
        Instruction(braket_gates.Rx(a * np.pi), q1),
        Instruction(braket_gates.Ry(b * np.pi), q2),
        Instruction(braket_gates.CNot(), [q2, q1]),
        Instruction(braket_gates.Rx(-0.5 * np.pi), q2),
        Instruction(braket_gates.Rz(c * np.pi), q2),
        Instruction(braket_gates.CNot(), [q1, q2]),
        *_translate_one_qubit_cirq_operation_to_braket_instruction(B1, q1),
        *_translate_one_qubit_cirq_operation_to_braket_instruction(B2, q2),
    ]



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tools for sampling from the noisy representations of ideal operations."""

import warnings
from copy import deepcopy
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import cirq
import numpy as np

from mitiq import QPROGRAM
from mitiq.interface import (
    accept_qprogram_and_validate,
    convert_from_mitiq,
    convert_to_mitiq,
)
from mitiq.pec.types import OperationRepresentation
from mitiq.utils import _equal


def sample_sequence(
    ideal_operation: QPROGRAM,
    representations: Sequence[OperationRepresentation],
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    num_samples: int = 1,
) -> Tuple[List[QPROGRAM], List[int], float]:
    """Samples a list of implementable sequences from the quasi-probability
    representation of the input ideal operation.
    Returns the list of sequences, the corresponding list of signs and the
    one-norm of the quasi-probability representation (of the input operation).

    For example, if the ideal operation is U with representation U = a A + b B,
    then this function returns A with probability :math:`|a| / (|a| + |b|)` and
    B with probability :math:`|b| / (|a| + |b|)`. Also returns sign(a)
    (sign(b)) and :math:`|a| + |b|` if A (B) is sampled.

    Note that the ideal operation can be a sequence of operations (circuit),
    for instance U = V W, as long as a representation is known. Similarly, A
    and B can be sequences of operations (circuits) or just single operations.

    Args:
        ideal_operation: The ideal operation from which an implementable
            sequence is sampled.
        representations: A list of representations of ideal operations in a
            noisy basis. If no representation is found for `ideal_operation`,
            a ValueError is raised.
        random_state: Seed for sampling.
        num_samples: The number of samples.

    Returns:
        The tuple (``sequences``, ``signs``, ``norm``) where
        ``sequences`` are the sampled sequences,
        ``signs`` are the signs associated to the sampled ``sequences`` and
        ``norm`` is the one-norm of the quasi-probability distribution.

    Raises:
        ValueError: If no representation is found for `ideal_operation`.
    """
    # Grab the representation for the given ideal operation.
    ideal, native_type = convert_to_mitiq(ideal_operation)
    operation_representation = None
    for representation in representations:
        if _equal(
            representation.ideal,
            ideal,
            require_qubit_equality=representation.is_qubit_dependent,
        ):
            operation_representation = representation
            break
    if operation_representation is None:
        warnings.warn(
            UserWarning(f"No representation found for \n\n{ideal_operation}.")
        )
        return (
            [ideal_operation] * num_samples,
            [1] * num_samples,
            1.0,
        )

    # Qubit mapping is necessary for qubit-independent operation reps
    qubit_map = dict(
        zip(
            sorted(operation_representation.ideal.all_qubits()),
            sorted(ideal.all_qubits()),
        )
    )

    # Sample from this representation.
    norm = operation_representation.norm
    sequences = []
    signs = []
    for _ in range(num_samples):
        noisy_op, sign, _ = operation_representation.sample(random_state)
        if operation_representation.is_qubit_dependent:
            native_circ = noisy_op.native_circuit
        else:
            cirq_circ = noisy_op.circuit
            cirq_circ = cirq_circ.transform_qubits(qubit_map)
            native_circ = convert_from_mitiq(cirq_circ, native_type)
        sequences.append(native_circ)
        signs.append(sign)

    return sequences, signs, norm


def _cirq_sample_circuit(
    ideal_circuit: cirq.Circuit,
    representations: Sequence[OperationRepresentation],
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    num_samples: int = 1,
    extra_data: Dict[str, Any] = {},
) -> List[cirq.Circuit]:
    """Cirq version of the more general "sample_circuit" function.

    Args:
        ideal_circuit: The ideal circuit from which an implementable circuit
            is sampled.
        representations: List of representations of every operation in the
            input circuit. If a representation cannot be found for an operation
            in the circuit, a ValueError is raised.
        random_state: Seed for sampling.
        num_samples: The number of samples,
        extra_data: Mutable dictionary to save extra data beyond output
            circuits. After the function call it will contain ``signs`` (the
            signs associated to sampled_circuits) and ``norm`` (the one-norm
            of the full circuit representation).

    Returns:
        The sampled implementable circuits.

    Raises:
        ValueError:
            If a representation is not found for an operation in the circuit.
    """
    if isinstance(random_state, int):
        random_state = np.random.RandomState(random_state)

    # copy and remove all moments
    sampled_circuits = [
        deepcopy(ideal_circuit)[0:0] for _ in range(num_samples)
    ]
    sampled_signs = [1 for _ in range(num_samples)]
    norm = 1.0

    for op in ideal_circuit.all_operations():
        sequences, loc_signs, loc_norm = sample_sequence(
            cirq.Circuit(op),
            representations,
            num_samples=num_samples,
            random_state=random_state,
        )
        norm *= loc_norm
        for j in range(num_samples):
            sampled_signs[j] *= loc_signs[j]
            cirq_seq, _ = convert_to_mitiq(sequences[j])
            sampled_circuits[j].append(cirq_seq.all_operations())

    extra_data.update({"signs": sampled_signs, "norm": norm})

    return sampled_circuits


def sample_circuit(
    ideal_circuit: QPROGRAM,
    representations: Sequence[OperationRepresentation],
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    num_samples: int = 1,
) -> Tuple[List[QPROGRAM], List[int], float]:
    """Samples a list of implementable circuits from the quasi-probability
    representation of the input ideal circuit.
    Returns the list of circuits, the corresponding list of signs and the
    one-norm of the quasi-probability representation (of the full circuit).

    Args:
        ideal_circuit: The ideal circuit from which an implementable circuit
            is sampled.
        representations: List of representations of every operation in the
            input circuit. If a representation cannot be found for an operation
            in the circuit, a ValueError is raised.
        random_state: Seed for sampling.
        num_samples: The number of samples.

    Returns:
        The tuple (``sampled_circuits``, ``signs``, ``norm``) where
        ``sampled_circuits`` are the sampled implementable circuits,
        ``signs`` are the signs associated to sampled_circuits and
        ``norm`` is the one-norm of the circuit representation.

    Raises:
        ValueError:
            If a representation is not found for an operation in the circuit.
    """
    qprogram_sample_circuit = accept_qprogram_and_validate(
        _cirq_sample_circuit,
        one_to_many=True,
    )
    extra_data: Dict[str, Any] = {}
    sampled_qprograms = qprogram_sample_circuit(
        ideal_circuit, representations, random_state, num_samples, extra_data
    )
    signs: List[int] = extra_data["signs"]
    norm: float = extra_data["norm"]

    return sampled_qprograms, signs, norm



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.


"""High-level Quantum Susbapce Expansion tools."""

from functools import wraps
from typing import Callable, Dict, List, Sequence, Union

from mitiq import QPROGRAM, Executor, Observable, PauliString, QuantumResult
from mitiq.qse.qse_utils import (
    get_expectation_value_for_observable,
    get_projector,
)


def execute_with_qse(
    circuit: QPROGRAM,
    executor: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    check_operators: Sequence[PauliString],
    code_hamiltonian: Observable,
    observable: Observable,
    pauli_string_to_expectation_cache: Dict[PauliString, complex] = {},
) -> float:
    """Function for the calculation of an observable from some circuit of
    interest to be mitigated with quantum subspace expansion (QSE).

    Args:
        circuit: Quantum program to execute with error mitigation.
        executor: Executes a circuit and returns a `QuantumResult`.
        check_operators: List of check operators that define the
            stabilizer code space.
        code_hamiltonian: Hamiltonian of the code space.
        observable: Observable to compute the mitigated expectation value of.
        pauli_string_to_expectation_cache: Cache for expectation values of
            Pauli strings used to compute the projector and the observable.

    Returns:
        The expectation value estimated with QSE.
    """
    projector = get_projector(
        circuit,
        executor,
        check_operators,
        code_hamiltonian,
        pauli_string_to_expectation_cache,
    )
    # Compute the expectation value of the observable: <P O P>
    pop = get_expectation_value_for_observable(
        circuit,
        executor,
        projector * observable * projector,
        pauli_string_to_expectation_cache,
    )
    # Compute the normalization factor: <P P>
    pp = get_expectation_value_for_observable(
        circuit,
        executor,
        projector * projector,
        pauli_string_to_expectation_cache,
    )
    return pop / pp


def mitigate_executor(
    executor: Callable[[QPROGRAM], QuantumResult],
    check_operators: Sequence[PauliString],
    code_hamiltonian: Observable,
    observable: Observable,
    pauli_string_to_expectation_cache: Dict[PauliString, complex] = {},
) -> Callable[[QPROGRAM], float]:
    """Returns a modified version of the input 'executor' which is
    error-mitigated with quantum subspace expansion (QSE).

    Args:
        executor: Executes a circuit and returns a `QuantumResult`.
        check_operators: List of check operators that define the
            stabilizer code space.
        code_hamiltonian: Hamiltonian of the code space.
        observable: Observable to compute the mitigated expectation value for.
        pauli_string_to_expectation_cache: Cache for expectation values of
            Pauli strings used to compute the projector and the observable.

    Returns:
        The error-mitigated version of the input executor.
    """
    executor_obj = Executor(executor)
    if not executor_obj.can_batch:

        @wraps(executor)
        def new_executor(circuit: QPROGRAM) -> float:
            return execute_with_qse(
                circuit,
                executor,
                check_operators,
                code_hamiltonian,
                observable,
                pauli_string_to_expectation_cache,
            )

    else:

        @wraps(executor)
        def new_executor(circuits: List[QPROGRAM]) -> List[float]:
            return [
                execute_with_qse(
                    circuit,
                    executor,
                    check_operators,
                    code_hamiltonian,
                    observable,
                    pauli_string_to_expectation_cache,
                )
                for circuit in circuits
            ]

    return new_executor


def qse_decorator(
    check_operators: Sequence[PauliString],
    code_hamiltonian: Observable,
    observable: Observable,
    pauli_string_to_expectation_cache: Dict[PauliString, complex] = {},
) -> Callable[
    [Callable[[QPROGRAM], QuantumResult]], Callable[[QPROGRAM], float]
]:
    """Decorator which adds an error-mitigation layer based on quantum
    subspace expansion (QSE) to an executor function, i.e., a function which
    executes a quantum circuit with an arbitrary backend and returns a
    ``QuantumResult``.

    Args:
        check_operators: List of check operators that define the
            stabilizer code space.
        code_hamiltonian: Hamiltonian of the code space.
        observable: Observable to compute the mitigated expectation value of.
        pauli_string_to_expectation_cache: Cache for expectation values of
            Pauli strings used to compute the projector and the observable.

    Returns:
        The error-mitigating decorator to be applied to an executor function.
    """

    def decorator(
        executor: Callable[[QPROGRAM], QuantumResult],
    ) -> Callable[[QPROGRAM], float]:
        val = mitigate_executor(
            executor,
            check_operators,
            code_hamiltonian,
            observable,
            pauli_string_to_expectation_cache,
        )
        return val

    return decorator



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from mitiq.qse.qse import execute_with_qse, mitigate_executor, qse_decorator
from mitiq.qse.qse_utils import (
    get_projector,
    get_expectation_value_for_observable,
)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions for creating mirror circuits as defined in
:cite:`Proctor_2021_NatPhys` for benchmarking quantum computers
(with error mitigation)."""

from typing import List, Optional, Tuple

import cirq
import networkx as nx
from cirq.experiments.qubit_characterizations import _single_qubit_cliffords
from numpy import random

from mitiq import QPROGRAM, Bitstring
from mitiq.interface import convert_from_mitiq

single_q_cliffords = _single_qubit_cliffords()
cliffords = single_q_cliffords.c1_in_xy
paulis = [cirq.X, cirq.Y, cirq.Z, cirq.I]


def random_paulis(
    connectivity_graph: nx.Graph, random_state: random.RandomState
) -> cirq.Circuit:
    """Returns a circuit with randomly selected Pauli gates on each qubit.

    Args:
        connectivity_graph: Connectivity graph of device to run circuit on.
        random_state: Random state to select Paulis I, X, Y, Z uniformly at
            random.
    """
    return cirq.Circuit(
        paulis[random_state.randint(len(paulis))](cirq.LineQubit(x))
        for x in connectivity_graph.nodes
    )


def edge_grab(
    two_qubit_gate_prob: float,
    connectivity_graph: nx.Graph,
    random_state: random.RandomState,
) -> nx.Graph:
    """
    Args:
        two_qubit_gate_prob: Probability of an edge being chosen
            from the set of candidate edges.
        connectivity_graph: The connectivity graph for the backend
            on which the circuit will be run.
        random_state: Random state to select edges (uniformly at random).

    Returns:
        Returns a set of edges for which two qubit gates are to be
        applied given a two qubit gate density and the connectivity graph
        that must be satisfied.
    """
    connectivity_graph = connectivity_graph.copy()
    candidate_edges = nx.Graph()

    final_edges = nx.Graph()
    final_edges.add_nodes_from(connectivity_graph)

    while connectivity_graph.edges:
        num = random_state.randint(connectivity_graph.size())
        edges = list(connectivity_graph.edges)
        curr_edge = edges[num]
        candidate_edges.add_edge(*curr_edge)
        connectivity_graph.remove_nodes_from(curr_edge)

    for edge in candidate_edges.edges:
        if random_state.uniform(0.0, 1.0) < two_qubit_gate_prob:
            final_edges.add_edge(*edge)
    return final_edges


def random_cliffords(
    connectivity_graph: nx.Graph,
    random_state: random.RandomState,
    two_qubit_gate: cirq.Gate = cirq.CNOT,
) -> cirq.Circuit:
    """
    Args:
        connectivity_graph: A graph with the edges for which the
            two-qubit Clifford gate is to be applied.
        random_state: Random state to choose Cliffords (uniformly at random).
        two_qubit_gate: Two-qubit gate to use.

    Returns:
        A circuit with a two-qubit Clifford gate applied to each edge in
        edges, and a random single-qubit Clifford gate applied to every
        other qubit.
    """
    gates = [
        two_qubit_gate.on(cirq.LineQubit(a), cirq.LineQubit(b))
        for a, b in list(connectivity_graph.edges)
    ]
    qubits = nx.Graph()
    qubits.add_nodes_from(nx.isolates(connectivity_graph))
    gates.extend(
        list(random_single_cliffords(qubits, random_state).all_operations())
    )
    return cirq.Circuit(gates)


def random_single_cliffords(
    connectivity_graph: nx.Graph, random_state: random.RandomState
) -> cirq.Circuit:
    """
    Args:
        connectivity_graph: A graph with each node representing a qubit for
            which a random single-qubit Clifford gate is to be applied.
        random_state: Random state to choose Cliffords (uniformly at random).

    Returns:
        A circuit with a random single-qubit Clifford gate applied on each
        given qubit.
    """
    gates: List[cirq.Operation] = []
    for qubit in connectivity_graph.nodes:
        num = random_state.randint(len(cliffords))
        for clifford_gate in cliffords[num]:
            gates.append(clifford_gate(cirq.LineQubit(qubit)))
    return cirq.Circuit(gates)


def generate_mirror_circuit(
    nlayers: int,
    two_qubit_gate_prob: float,
    connectivity_graph: nx.Graph,
    two_qubit_gate_name: str = "CNOT",
    seed: Optional[int] = None,
    return_type: Optional[str] = None,
) -> Tuple[QPROGRAM, Bitstring]:
    """
    Args:
        nlayers: The number of random Clifford layers to be generated.
        two_qubit_gate_prob: Probability of a two-qubit gate being applied.
        connectivity_graph: The connectivity graph of the backend
            on which the mirror circuit will be run. This is used
            to make sure 2-qubit gates are only applied to connected qubits.
        two_qubit_gate_name: Name of two-qubit gate to use. Options are "CNOT"
            and "CZ".
        seed: Seed for generating randomized mirror circuit.
        return_type: String which specifies the type of the
            returned circuit. See the keys of ``mitiq.SUPPORTED_PROGRAM_TYPES``
            for options. If ``None``, the returned circuit is a
            ``cirq.Circuit``.

    Returns:
        A randomized mirror circuit and the bitstring corresponding
        to a noise free result.
    """
    if not 0 <= two_qubit_gate_prob <= 1:
        raise ValueError("two_qubit_gate_prob must be between 0 and 1")

    supported_two_qubit_gates = {"CZ": cirq.CZ, "CNOT": cirq.CNOT}
    if two_qubit_gate_name not in supported_two_qubit_gates.keys():
        raise ValueError(
            f"Supported two-qubit gate names are "
            f"{tuple(supported_two_qubit_gates.keys())} but "
            f"{two_qubit_gate_name} was provided for `two_qubit_gate_name`."
        )
    two_qubit_gate = supported_two_qubit_gates[two_qubit_gate_name]

    random_state = random.RandomState(seed)

    single_qubit_cliffords = random_single_cliffords(
        connectivity_graph, random_state=random_state
    )

    forward_circuit = cirq.Circuit()

    quasi_inversion_circuit = cirq.Circuit()
    quasi_inverse_gates = []

    for _ in range(nlayers):
        forward_circuit.append(random_paulis(connectivity_graph, random_state))

        selected_edges = edge_grab(
            two_qubit_gate_prob, connectivity_graph, random_state
        )
        circ = random_cliffords(selected_edges, random_state, two_qubit_gate)
        forward_circuit.append(circ)

        quasi_inverse_gates.append(
            random_paulis(connectivity_graph, random_state)
        )
        quasi_inverse_gates.append(cirq.inverse(circ))

    quasi_inversion_circuit.append(
        gate for gate in reversed(quasi_inverse_gates)
    )

    rand_paulis = cirq.Circuit(random_paulis(connectivity_graph, random_state))
    circuit = (
        single_qubit_cliffords
        + forward_circuit
        + rand_paulis
        + quasi_inversion_circuit
        + cirq.inverse(single_qubit_cliffords)
    )

    # Compute the bitstring this circuit should sample.
    res = cirq.Simulator().run(
        circuit + cirq.measure(*sorted(circuit.all_qubits()))
    )
    bitstring = list(res.measurements.values())[0][0].tolist()

    return_type = "cirq" if not return_type else return_type
    return convert_from_mitiq(circuit, return_type), bitstring



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions for local and global unitary folding on supported circuits."""

import warnings
from copy import deepcopy
from typing import Any, Dict, FrozenSet, List, Optional, cast

import numpy as np
from cirq import Circuit, InsertStrategy, Moment, has_unitary, inverse, ops

from mitiq.interface import accept_qprogram_and_validate
from mitiq.utils import (
    _append_measurements,
    _is_measurement,
    _pop_measurements,
)


class UnfoldableCircuitError(Exception):
    pass


_cirq_gates_to_string_keys = {
    ops.H: "H",
    ops.X: "X",
    ops.Y: "Y",
    ops.Z: "Z",
    ops.T: "T",
    ops.I: "I",
    ops.S: "S",
    ops.T: "T",
    ops.rx: "rx",
    ops.ry: "ry",
    ops.rz: "rz",
    ops.CNOT: "CNOT",
    ops.CZ: "CZ",
    ops.SWAP: "SWAP",
    ops.ISWAP: "ISWAP",
    ops.CSWAP: "CSWAP",
    ops.TOFFOLI: "TOFFOLI",
}
_string_keys_to_cirq_gates = {
    opstring: op for op, opstring in _cirq_gates_to_string_keys.items()
}

_valid_gate_names = list(
    map(
        lambda gate_name: gate_name.lower(),
        _cirq_gates_to_string_keys.values(),
    )
) + ["single", "double", "triple"]


# Helper functions
def _check_foldable(circuit: Circuit) -> None:
    """Raises an error if the input circuit cannot be folded.

    Args:
        circuit: Checks whether this circuit is able to be folded.

    Raises:
        UnfoldableCircuitError:
            * If the circuit has intermediate measurements.
            * If the circuit has non-unitary channels which are not terminal
              measurements.
    """
    if not circuit.are_all_measurements_terminal():
        raise UnfoldableCircuitError(
            "Circuit contains intermediate measurements and cannot be folded."
        )

    if not has_unitary(circuit):
        circ_measurements = _pop_measurements(circuit)
        if inverse(circuit, default=None) is None:
            raise UnfoldableCircuitError(
                "Circuit contains non-invertible channels which are not"
                "terminal measurements and cannot be folded."
            )
        else:
            _append_measurements(circuit, circ_measurements)


def _squash_moments(circuit: Circuit) -> Circuit:
    """Returns a copy of the input circuit with all gates squashed into as few
    moments as possible.

    Args:
        circuit: Circuit to squash moments of.
    """
    output_circuit = circuit.copy()
    output_circuit = output_circuit[0:0]  # Remove moments
    output_circuit.append(
        circuit.all_operations(), strategy=InsertStrategy.EARLIEST
    )
    return output_circuit


def _fold_all(
    circuit: Circuit,
    num_folds: int = 1,
    exclude: FrozenSet[Any] = frozenset(),
    skip_moments: FrozenSet[int] = frozenset(),
) -> Circuit:
    """Returns a circuit with all gates folded locally.

    Args:
        circuit: Circuit to fold.
        num_folds: Number of times to add G G^dag for each gate G. If not an
            integer, this gets rounded to the nearest integer.
        exclude: Do not fold these gates.
        skip_moments: Do not fold these moments.
    """
    num_folds = round(num_folds)
    if num_folds < 0:
        raise ValueError(
            f"Arg `num_folds` must be positive but was {num_folds}."
        )

    # Parse the exclude argument.
    all_gates = set(cast(ops.Gate, op.gate) for op in circuit.all_operations())
    to_exclude = set()
    for item in exclude:
        if isinstance(item, str):
            try:
                to_exclude.add(_string_keys_to_cirq_gates[item])
            except KeyError:
                if item == "single":
                    to_exclude.update(
                        gate for gate in all_gates if gate.num_qubits() == 1
                    )
                elif item == "double":
                    to_exclude.update(
                        gate for gate in all_gates if gate.num_qubits() == 2
                    )
                elif item == "triple":
                    to_exclude.update(
                        gate for gate in all_gates if gate.num_qubits() == 3
                    )
                else:
                    raise ValueError(
                        f"Do not know how to parse item '{item}' in exclude. "
                        f"Valid items are Cirq gates, string keys specifying"
                        f"gates, and 'single', 'double', or 'triple'."
                    )
        elif isinstance(item, ops.Gate):
            to_exclude.add(item)
        else:
            raise ValueError(
                f"Do not know how to exclude {item} of type {type(item)}."
            )

    folded = deepcopy(circuit)[:0]
    for i, moment in enumerate(circuit):
        if i in skip_moments:
            folded.append(moment, strategy=InsertStrategy.EARLIEST)
            continue

        for op in moment:
            folded.append(op, strategy=InsertStrategy.EARLIEST)
            if op.gate not in to_exclude:
                folded.append(
                    [inverse(op), op] * num_folds,
                    strategy=InsertStrategy.EARLIEST,
                )

    return folded


# Helper functions for folding by fidelity
def _default_weight(op: ops.Operation) -> float:
    """Returns a default weight for an operation."""
    return 0.99 ** len(op.qubits)


def _get_weight_for_gate(
    weights: Dict[str, float], op: ops.Operation
) -> float:
    """Returns the weight for a given gate, using a default value of 1.0 if
    weights is None or if the weight is not specified.

    Args:
        weights: Dictionary of string keys mapping gates to weights.
        op: Operation to get the weight of.
    """
    weight = _default_weight(op)
    if not weights:
        return weight

    if "single" in weights.keys() and len(op.qubits) == 1:
        weight = weights["single"]
    elif "double" in weights.keys() and len(op.qubits) == 2:
        weight = weights["double"]
    elif "triple" in weights.keys() and len(op.qubits) == 3:
        weight = weights["triple"]

    if op.gate and op.gate in _cirq_gates_to_string_keys.keys():
        # Get the string key for this gate
        key = _cirq_gates_to_string_keys[op.gate]
        if key in weights.keys():
            weight = weights[_cirq_gates_to_string_keys[op.gate]]
    return weight


# Local folding functions
@accept_qprogram_and_validate
def fold_all(
    circuit: Circuit,
    scale_factor: float,
    exclude: FrozenSet[Any] = frozenset(),
) -> Circuit:
    """Returns a circuit with all gates folded locally.

    Args:
        circuit: Circuit to fold.
        scale_factor: Approximate factor by which noise is scaled in the
            circuit. Each gate is folded round((scale_factor - 1.0) / 2.0)
            times. For example::

                scale_factor | num_folds
                ------------------------
                1.0          | 0
                3.0          | 1
                5.0          | 2

        exclude: Do not fold these gates. Supported gate keys are listed in
            the following table.::

                Gate key    | Gate
                -------------------------
                "H"         | Hadamard
                "X"         | Pauli X
                "Y"         | Pauli Y
                "Z"         | Pauli Z
                "I"         | Identity
                "S"         | Phase gate
                "T"         | T gate
                "rx"        | X-rotation
                "ry"        | Y-rotation
                "rz"        | Z-rotation
                "CNOT"      | CNOT
                "CZ"        | CZ gate
                "SWAP"      | Swap
                "ISWAP"     | Imaginary swap
                "CSWAP"     | CSWAP
                "TOFFOLI"   | Toffoli gate
                "single"    | All single qubit gates
                "double"    | All two-qubit gates
                "triple"    | All three-qubit gates
    """
    if not 1.0 <= scale_factor:
        raise ValueError(
            f"Requires scale_factor >= 1 but scale_factor = {scale_factor}."
        )
    _check_foldable(circuit)

    folded = deepcopy(circuit)
    measurements = _pop_measurements(folded)

    folded = _fold_all(folded, round((scale_factor - 1.0) / 2.0), exclude)

    _append_measurements(folded, measurements)
    return folded


# Global folding function
@accept_qprogram_and_validate
def fold_global(
    circuit: Circuit, scale_factor: float, **kwargs: Any
) -> Circuit:
    """Returns a new circuit obtained by folding the global unitary of the
    input circuit.

    The returned folded circuit has a number of gates approximately equal to
    scale_factor * len(circuit).

    Args:
        circuit: Circuit to fold.
        scale_factor: Factor to scale the circuit by.

    Keyword Args:
        return_mitiq (bool): If True, returns a Mitiq circuit instead of
            the input circuit type (if different). Default is False.

    Returns:
        folded: the folded quantum circuit as a QPROGRAM.
    """
    _check_foldable(circuit)

    if not (scale_factor >= 1):
        raise ValueError("The scale factor must be a real number >= 1.")

    folded = deepcopy(circuit)
    measurements = _pop_measurements(folded)
    base_circuit = deepcopy(folded)

    # Determine the number of global folds and the final fractional scale
    num_global_folds, fraction_scale = divmod(scale_factor - 1, 2)
    # Do the global folds
    for _ in range(int(num_global_folds)):
        folded += Circuit(inverse(base_circuit), base_circuit)

    # Fold remaining gates until the scale is reached
    operations = list(base_circuit.all_operations())
    num_to_fold = int(round(fraction_scale * len(operations) / 2))

    if num_to_fold > 0:
        # Create the inverse of the final partial circuit
        inverse_partial = Circuit()
        num_partial = 0
        for moment in base_circuit[::-1]:
            new_moment = Moment()
            for op in moment.operations[::-1]:
                new_moment = new_moment.with_operation(inverse(op))
                num_partial += 1
                if num_partial == num_to_fold:
                    break
            inverse_partial.append(new_moment)
            if num_partial == num_to_fold:
                break
        # Append partially folded circuit
        folded += inverse_partial + inverse(inverse_partial)

    _append_measurements(folded, measurements)
    return folded


def _create_weight_mask(
    circuit: Circuit,
    fidelities: Dict[str, float],
) -> List[float]:
    """Returns a list of weights associated to each gate in the input
    circuit. Measurement gates are ignored.

    The gate ordering is equal to the one used in the `all_operations()`
    method of the :class:`cirq.Circuit` class: gates from earlier moments
    come first and gates within the same moment follow the order in which
    they were given to the moment's constructor.

    Args:
        circuit: The circuit from which a weight mask is created.
        fidelities: The dictionary of gate fidelities.
            If None, default fidelities will be used. See the
            docstring of local folding function for mode details.

    Returns: The list of weights associated to all the gates.
    """
    if not all(0.0 < f <= 1.0 for f in fidelities.values()):
        raise ValueError("Fidelities should be in the interval (0, 1].")

    invalid_fidelities = filter(
        lambda opname: opname.lower() not in _valid_gate_names, fidelities
    )
    for gate_name in invalid_fidelities:
        warnings.warn(
            f"You passed a fidelity for the gate '{gate_name}', but we don't "
            "currently support this gate."
        )

    # Round to avoid ugly numbers like 0.09999999999999998 instead of 0.1
    weights = {k: round(1.0 - f, 12) for k, f in fidelities.items()}

    # Build mask with weights of each gate
    return [
        _get_weight_for_gate(weights, op)
        for op in circuit.all_operations()
        if not _is_measurement(op)
    ]


def _create_fold_mask(
    weight_mask: List[float],
    scale_factor: float,
    seed: Optional[int] = None,
) -> List[int]:
    r"""Returns a list of integers determining how many times each gate a
    circuit should be folded to realize the desired input scale_factor.

    More precicely, the j_th element of the returned list is associated
    to the j_th gate G_j of a circuit that we want to scale and determines
    how many times G_j^\dag G_j should be applied after G_j.

    The gate ordering is equal to the one used in the `all_operations()`
    method of the :class:`cirq.Circuit` class: gates from earlier moments
    come first and gates within the same moment follow the order in which
    they were given to the moment's constructor.

    The returned list is built such that the total weight of the
    folded circuit is approximately equal to scale_factor times the
    total weight of the input circuit.

    For equal weights, this function reproduces the local unitary folding
    method defined in equation (5) of :cite:`Giurgica_Tiron_2020_arXiv`, with
    the layer indices chosen at random.

    Args:
        weight_mask: The weights of all the gates of the circuit to fold.
            Highly noisy gates should have a corresponding high weight.
            Gates with zero weight are assumed to be ideal and are not folded.
        scale_factor: The effective noise scale factor.
        seed: A seed for the random number generator used to select the
            subset of layer indices to fold.

    Returns: The list of integers determining to how many times one should
        fold the j_th gate of the circuit to be scaled.

    Example:
        >>>_create_fold_mask(
            weight_mask=[1.0, 0.5, 2.0, 0.0],
            scale_factor=4,
        )
        [2, 2, 1, 0]
    """

    if not 1.0 <= scale_factor:
        raise ValueError(
            f"Requires scale_factor >= 1 but scale_factor = {scale_factor}."
        )

    # Find the maximum odd integer smaller or equal to scale_factor
    num_uniform_folds = int((scale_factor - 1.0) / 2.0)
    odd_integer_scale_factor = 2 * num_uniform_folds + 1

    # Uniformly folding all gates to reach odd_integer_scale_factor
    num_folds_mask = []
    for w in weight_mask:
        if np.isclose(w, 0.0):
            num_folds_mask.append(0)
        else:
            num_folds_mask.append(num_uniform_folds)

    # If the scale_factor is an odd integer, we are done.
    if np.isclose(odd_integer_scale_factor, scale_factor):
        return num_folds_mask

    # Express folding order through a list of indices
    folding_order = list(range(len(weight_mask)))

    rnd_state = np.random.RandomState(seed)
    rnd_state.shuffle(folding_order)

    # Fold gates until the input scale_factor is better approximated
    input_circuit_weight = sum(weight_mask)
    output_circuit_weight = odd_integer_scale_factor * input_circuit_weight
    approx_error = np.abs(
        output_circuit_weight - scale_factor * input_circuit_weight
    )
    for j in folding_order:
        # Skip gates with 0 weight
        if np.isclose(weight_mask[j], 0.0):
            continue
        # Compute the approx error if a new fold would be applied
        new_output_circuit_weight = output_circuit_weight + 2 * weight_mask[j]
        new_approx_error = np.abs(
            new_output_circuit_weight - scale_factor * input_circuit_weight
        )
        # Fold the candidate gate only if it helps improving the approximation
        if new_approx_error >= approx_error:
            break
        approx_error = new_approx_error
        output_circuit_weight = new_output_circuit_weight
        num_folds_mask[j] += 1

    return num_folds_mask


def _apply_fold_mask(
    circuit: Circuit,
    num_folds_mask: List[int],
    squash_moments: Optional[bool] = True,
) -> Circuit:
    r"""Applies local unitary folding to the gates of the input circuit
    according to the input num_folds_mask.

    More precicely, G_j^\dag G_j is applied after the j_th gate G_j of
    the input circuit an integer number of times given by num_folds_mask[j].

    The gate ordering is equal to the one used in the `all_operations()`
    method of the :class:`cirq.Circuit` class: gates from earlier moments
    come first and gates within the same moment follow the order in which
    they were given to the moment's constructor.

    Args:
        circuit: The quantum circuit to fold.
        num_folds_mask: The list of integers indicating how many times
            the corresponding gates of 'circuit' should be folded.
        squash_moments: If True or None, all gates (including folded gates) are
            placed as early as possible in the circuit. If False, new moments
            are created for folded gates. This option only applies to QPROGRAM
            types which have a "moment" or "time" structure. Default is True.

    Returns: The folded quantum circuit.
    """
    _check_foldable(circuit)
    circ_copy = deepcopy(circuit)
    measurements = _pop_measurements(circ_copy)

    num_gates = len(list(circ_copy.all_operations()))
    if num_gates != len(num_folds_mask):
        raise ValueError(
            "The circuit and the folding mask have incompatible sizes."
            f" The number of gates is {num_gates}"
            f" but len(num_folds_mask) is {len(num_folds_mask)}."
        )

    folded_circuit = circ_copy[:0]
    if squash_moments:
        for op, num_folds in zip(circ_copy.all_operations(), num_folds_mask):
            folded_circuit.append([op] + num_folds * [inverse(op), op])
    else:
        mask_index = 0
        for moment in circ_copy:
            folded_moment = Circuit(moment)
            for op in moment:
                num_folds = num_folds_mask[mask_index]
                folded_moment.append(num_folds * [inverse(op), op])
                mask_index += 1
            # New folded gates are only squashed with respect to folded_moment
            # while folded_circuit is not squashed.
            folded_circuit.append(folded_moment)

    _append_measurements(folded_circuit, measurements)
    return folded_circuit


@accept_qprogram_and_validate
def fold_gates_at_random(
    circuit: Circuit,
    scale_factor: float,
    seed: Optional[int] = None,
    **kwargs: Any,
) -> Circuit:
    r"""
    Returns a new folded circuit by applying the map G -> G G^dag G to a
    subset of gates of the input circuit, different indices randomly sampled
    without replacement.

    The folded circuit has a number of gates approximately equal to
    scale_factor * n where n is the number of gates in the input circuit.

    For equal gate fidelities, this function reproduces the local unitary
    folding method defined in equation (5) of
    :cite:`Giurgica_Tiron_2020_arXiv`.

    Args:
        circuit: Circuit to fold.
        scale_factor: Factor to scale the circuit by. Any real number >= 1.
        seed: Seed for random number generator.

    Keyword Args:
        fidelities (Dict[str, float]): Dictionary of gate fidelities. Each key
            is a string which specifies the gate and each value is the
            fidelity of that gate. When this argument is provided, folded
            gates contribute an amount proportional to their infidelity
            (1 - fidelity) to the total noise scaling. Fidelity values must be
            in the interval (0, 1]. Gates not specified have a default
            fidelity of 0.99**n where n is the number of qubits the gates act
            on.

            Supported gate keys are listed in the following table.::

                Gate key    | Gate
                -------------------------
                "H"         | Hadamard
                "X"         | Pauli X
                "Y"         | Pauli Y
                "Z"         | Pauli Z
                "I"         | Identity
                "S"         | Phase gate
                "T"         | T gate
                "rx"        | X-rotation
                "ry"        | Y-rotation
                "rz"        | Z-rotation
                "CNOT"      | CNOT
                "CZ"        | CZ gate
                "SWAP"      | Swap
                "ISWAP"     | Imaginary swap
                "CSWAP"     | CSWAP
                "TOFFOLI"   | Toffoli gate
                "single"    | All single qubit gates
                "double"    | All two-qubit gates
                "triple"    | All three-qubit gates

            Keys for specific gates override values set by "single", "double",
            and "triple".

            For example, `fidelities = {"single": 1.0, "H", 0.99}` sets all
            single-qubit gates except Hadamard to have fidelity one.

        squash_moments (bool): If True, all gates (including folded gates) are
            placed as early as possible in the circuit. If False, new moments
            are created for folded gates. This option only applies to QPROGRAM
            types which have a "moment" or "time" structure. Default is True.

        return_mitiq (bool): If True, returns a Mitiq circuit instead of
            the input circuit type (if different). Default is False.

    Returns:
        folded: The folded quantum circuit as a QPROGRAM.

    """

    weight_mask = _create_weight_mask(circuit, kwargs.get("fidelities", {}))

    num_folds_mask = _create_fold_mask(
        weight_mask,
        scale_factor,
        seed=seed,
    )

    return _apply_fold_mask(
        circuit,
        num_folds_mask,
        squash_moments=kwargs.get("squash_moments", True),
    )



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from mitiq.executor.executor import Executor



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import random
from functools import singledispatch
from typing import Callable, List, Optional

import cirq
import pennylane as qml
from cirq import Circuit as _Circuit
from pennylane.tape import QuantumTape

from mitiq import QPROGRAM
from mitiq.interface import accept_qprogram_and_validate

# P, Q, R, S from https://arxiv.org/pdf/2301.02690.pdf
CNOT_twirling_gates = [
    (cirq.I, cirq.I, cirq.I, cirq.I),
    (cirq.I, cirq.X, cirq.I, cirq.X),
    (cirq.I, cirq.Y, cirq.Z, cirq.Y),
    (cirq.I, cirq.Z, cirq.Z, cirq.Z),
    (cirq.Y, cirq.I, cirq.Y, cirq.X),
    (cirq.Y, cirq.X, cirq.Y, cirq.I),
    (cirq.Y, cirq.Y, cirq.X, cirq.Z),
    (cirq.Y, cirq.Z, cirq.X, cirq.Y),
    (cirq.X, cirq.I, cirq.X, cirq.X),
    (cirq.X, cirq.X, cirq.X, cirq.I),
    (cirq.X, cirq.Y, cirq.Y, cirq.Z),
    (cirq.X, cirq.Z, cirq.Y, cirq.Y),
    (cirq.Z, cirq.I, cirq.Z, cirq.I),
    (cirq.Z, cirq.X, cirq.Z, cirq.X),
    (cirq.Z, cirq.Y, cirq.I, cirq.Y),
    (cirq.Z, cirq.Z, cirq.I, cirq.Z),
]
CZ_twirling_gates = [
    (cirq.I, cirq.I, cirq.I, cirq.I),
    (cirq.I, cirq.X, cirq.Z, cirq.X),
    (cirq.I, cirq.Y, cirq.Z, cirq.Y),
    (cirq.I, cirq.Z, cirq.I, cirq.Z),
    (cirq.X, cirq.I, cirq.X, cirq.Z),
    (cirq.X, cirq.X, cirq.Y, cirq.Y),
    (cirq.X, cirq.Y, cirq.Y, cirq.X),
    (cirq.X, cirq.Z, cirq.X, cirq.I),
    (cirq.Y, cirq.I, cirq.Y, cirq.Z),
    (cirq.Y, cirq.X, cirq.X, cirq.Y),
    (cirq.Y, cirq.Y, cirq.X, cirq.X),
    (cirq.Y, cirq.Z, cirq.Y, cirq.I),
    (cirq.Z, cirq.I, cirq.Z, cirq.I),
    (cirq.Z, cirq.X, cirq.I, cirq.X),
    (cirq.Z, cirq.Y, cirq.I, cirq.Y),
    (cirq.Z, cirq.Z, cirq.Z, cirq.Z),
]

CIRQ_NOISE_FUNCTION = Callable[[float], cirq.Gate]

CIRQ_NOISE_OP: dict[str, CIRQ_NOISE_FUNCTION] = {
    "bit-flip": cirq.bit_flip,
    "depolarize": cirq.depolarize,
}

PENNYLANE_NOISE_OP = {
    "bit-flip": qml.BitFlip,
    "depolarize": qml.DepolarizingChannel,
}


def generate_pauli_twirl_variants(
    circuit: QPROGRAM,
    num_circuits: int = 10,
    noise_name: Optional[str] = None,
    **kwargs: float,
) -> List[QPROGRAM]:
    r"""Return the Pauli twirled versions of the input circuit.

    Only the CNOT and CZ gates in an input circuit are Pauli twirled
    as specified in :cite:`Saki_2023_arxiv`.

    Args:
        circuit: The input circuit on which twirling is applied.
        num_circuits: Number of twirled variants of the circuits.
        noise_name: Name of the noisy operator acting on CNOT and CZ gates.
            This is useful if the user requires a noisy circuit after twirling.
            Values allowed: ["bit-flip", "depolarize"]

    Returns:
        A list of `num_circuits` twirled versions of `circuit`
    """
    CNOT_twirled_circuits = twirl_CNOT_gates(circuit, num_circuits)
    twirled_circuits = [
        twirl_CZ_gates(c, num_circuits=1)[0] for c in CNOT_twirled_circuits
    ]

    if noise_name is not None:
        twirled_circuits = [
            add_noise_to_two_qubit_gates(circuit, noise_name, **kwargs)
            for circuit in twirled_circuits
        ]

    return twirled_circuits


def add_noise_to_two_qubit_gates(
    circuit: QPROGRAM, noise_name: str, **kwargs: float
) -> QPROGRAM:
    """Add noise to CNOT and CZ gates on pre-twirled circuits.

    Args:
        circuit: Pre-twirled circuit
        noise_name: name of noise operator to apply after CNOT and CZ gates
    """
    # here we will validate if noise_op and kwargs are valid
    return _add_noise_to_two_qubit_gates(circuit, noise_name, **kwargs)


@singledispatch
def _add_noise_to_two_qubit_gates(
    circuit: QPROGRAM, noise_name: str, **kwargs: float
) -> QPROGRAM:
    raise NotImplementedError(
        f"Cannot add noise to Circuit of type {type(circuit)}."
    )


@_add_noise_to_two_qubit_gates.register
def _cirq(circuit: _Circuit, noise_name: str, **kwargs: float) -> _Circuit:
    noise_function = CIRQ_NOISE_OP[noise_name]
    noise_op = noise_function(**kwargs)  # type: ignore

    noisy_gates = [cirq.ops.CNOT, cirq.ops.CZ]

    noisy_circuit = cirq.Circuit()
    for moment in circuit:
        layer = cirq.Circuit()
        for op in moment:
            layer.append(op)
            if op.gate in noisy_gates:
                layer.append(noise_op.on_each(op.qubits))
        noisy_circuit += layer
    return noisy_circuit


@_add_noise_to_two_qubit_gates.register
def _pennylane(
    circuit: QuantumTape, noise_name: str, **kwargs: float
) -> QuantumTape:
    new_ops = []
    noise_function = PENNYLANE_NOISE_OP[noise_name]

    noisy_gates = ["CNOT", "CZ"]
    for op in circuit:
        new_ops.append(op)
        if op.name in noisy_gates:
            for wire in op.wires:
                noise_op = noise_function(**kwargs, wires=wire)
                new_ops.append(noise_op)

    return QuantumTape(
        ops=new_ops, measurements=circuit.measurements, shots=circuit.shots
    )


def twirl_CNOT_gates(circuit: QPROGRAM, num_circuits: int) -> List[QPROGRAM]:
    """Generate a list of circuits using Pauli twirling on CNOT gates.

    Args:
        circuit: The circuit to generate twirled versions of
        num_circuits: The number of sampled circuits to return
    """
    return [_twirl_CNOT_qprogram(circuit) for _ in range(num_circuits)]


@accept_qprogram_and_validate
def _twirl_CNOT_qprogram(circuit: cirq.Circuit) -> cirq.Circuit:
    return circuit.map_operations(_twirl_single_CNOT_gate)


def twirl_CZ_gates(circuit: QPROGRAM, num_circuits: int) -> List[QPROGRAM]:
    """Generate a list of circuits using Pauli twirling on CZ gates.

    Args:
        circuit: The circuit to generate twirled versions of
        num_circuits: The number of sampled circuits to return
    """
    return [_twirl_CZ_qprogram(circuit) for _ in range(num_circuits)]


@accept_qprogram_and_validate
def _twirl_CZ_qprogram(circuit: cirq.Circuit) -> cirq.Circuit:
    return circuit.map_operations(_twirl_single_CZ_gate)


def _twirl_single_CNOT_gate(op: cirq.Operation) -> cirq.OP_TREE:
    """Function which converts a CNOT gate to a logical equivalent. That is,
    it converts CNOT operations into the following, leaving all other cirq
    operations alone.

        --P---⏺---R--
              |
        --Q---⊕---S--

    Where P, Q, R, and S are Pauli operators sampled so as to not effect
    change on the underlying unitary.
    """
    if op.gate != cirq.CNOT:
        return op

    P, Q, R, S = random.choice(CNOT_twirling_gates)
    control_qubit, target_qubit = op.qubits
    return [
        P.on(control_qubit),
        Q.on(target_qubit),
        op,
        R.on(control_qubit),
        S.on(target_qubit),
    ]


def _twirl_single_CZ_gate(op: cirq.Operation) -> cirq.OP_TREE:
    """Function which converts a CZ gate to a logical equivalent. That is, it
    converts CZ operations into the following, leaving all other cirq
    operations alone.

        --P---⏺---R--
              |
        --Q---⏺---S--

    Where P, Q, R, and S are Pauli operators sampled so as to not effect
    change on the underlying unitary.
    """
    if op.gate != cirq.CZ:
        return op

    P, Q, R, S = random.choice(CZ_twirling_gates)
    control_qubit, target_qubit = op.qubits
    return [
        P.on(control_qubit),
        Q.on(target_qubit),
        op,
        R.on(control_qubit),
        S.on(target_qubit),
    ]



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Provides a utility basic_compile to allow for running on QCS without
using quilc, as quilc will optimize away unitary folding.

NB: Copied in large part from rigetti/forest-benchmarking (Apache-2.0)
and modified to support a larger gateset (e.g. CPHASE).
"""

from typing import cast

import numpy as np
from pyquil.gates import CZ, RX, RZ, XY, I
from pyquil.quil import Program
from pyquil.quilatom import ParameterDesignator, QubitDesignator
from pyquil.quilbase import Gate


def _CCNOT(
    q0: QubitDesignator, q1: QubitDesignator, q2: QubitDesignator
) -> Program:
    """
    A CCNOT in terms of RX(+-pi/2), RZ(theta), and CZ

    .. note:
        Don't control this gate.
    """
    p = Program()
    p.inst(_H(q2))
    p.inst(_CNOT(q1, q2))
    p.inst(_T(q2, dagger=True))
    p.inst(_SWAP(q1, q2))
    p.inst(_CNOT(q0, q1))
    p.inst(_T(q1))
    p.inst(_CNOT(q2, q1))
    p.inst(_T(q1, dagger=True))
    p.inst(_CNOT(q0, q1))
    p.inst(_SWAP(q1, q2))
    p.inst(_T(q1))
    p.inst(_T(q2))
    p.inst(_CNOT(q0, q1))
    p.inst(_H(q2))
    p.inst(_T(q0))
    p.inst(_T(q1, dagger=True))
    p.inst(_CNOT(q0, q1))
    return p


def _CNOT(q0: QubitDesignator, q1: QubitDesignator) -> Program:
    """
    A CNOT in terms of RX(+-pi/2), RZ(theta), and CZ

    .. note:
        This uses two of :py:func:`_H`, so it picks up twice the global phase.
        Don't control this gate.
    """
    p = Program()
    p.inst(_H(q1))
    p.inst(CZ(q0, q1))
    p.inst(_H(q1))
    return p


def _CPHASE(
    angle: ParameterDesignator, q0: QubitDesignator, q1: QubitDesignator
) -> Program:
    """
    from quilc:

    (define-compiler CPHASE-to-CNOT ((CPHASE-gate ("CPHASE" (alpha) p q)))
        (inst "CNOT" ()                            p q)
        (inst "RZ"   (list (param-* alpha -0.5d0)) q)
        (inst "CNOT" ()                            p q)
        (inst "RZ"   (list (param-* alpha  0.5d0)) q)
        (inst "RZ"   (list (param-* alpha  0.5d0)) p))
    """
    p = Program()
    p.inst(_CNOT(q0, q1))
    p.inst(RZ(-0.5 * angle, q1))
    p.inst(_CNOT(q0, q1))
    p.inst(RZ(0.5 * angle, q1))
    p.inst(RZ(0.5 * angle, q0))
    return p


def _H(q: QubitDesignator) -> Program:
    """
    A Hadamard in terms of RX(+-pi/2) and RZ(theta)

    .. note:
        This introduces a different global phase! Don't control this gate.
    """
    p = Program()
    p.inst(_RY(-np.pi / 2, q))
    p.inst(RZ(np.pi, q))
    return p


def _ISWAP(q0: QubitDesignator, q1: QubitDesignator) -> Program:
    """
    An ISWAP as an XY(pi). Of course, assumes XY is available.
    """
    p = Program()
    p += XY(np.pi, q0, q1)
    return p


def _PHASE(angle: ParameterDesignator, q: QubitDesignator) -> Program:
    """
    from quilc:

    (define-compiler PHASE-to-RZ ((phase-gate ("PHASE" (alpha) q)))
       (inst "RZ" `(,alpha) q))
    """
    p = Program()
    p += RZ(angle, q)
    return p


def _RX(angle: ParameterDesignator, q: QubitDesignator) -> Program:
    """
    A RX in terms of native RX(+-pi/2) and RZ gates.
    """
    p = Program()
    p += RZ(np.pi / 2, q)
    p += RX(np.pi / 2, q)
    p += RZ(angle, q)
    p += RX(-np.pi / 2, q)
    p += RZ(-np.pi / 2, q)
    return p


def _RY(angle: ParameterDesignator, q: QubitDesignator) -> Program:
    """
    A RY in terms of RX(+-pi/2) and RZ(theta)
    """
    p = Program()
    p += RX(np.pi / 2, q)
    p += RZ(angle, q)
    p += RX(-np.pi / 2, q)
    return p


def _S(q: QubitDesignator) -> Program:
    """
    An S in terms of RZ(theta)
    """
    return Program(RZ(np.pi / 2, q))


def _SWAP(q0: QubitDesignator, q1: QubitDesignator) -> Program:
    """
    A SWAP in terms of _CNOT

    .. note:
        This uses :py:func:`_CNOT`, so it picks up a global phase.
        Don't control this gate.
    """
    p = Program()
    p.inst(_CNOT(q0, q1))
    p.inst(_CNOT(q1, q0))
    p.inst(_CNOT(q0, q1))
    return p


def _T(q: QubitDesignator, dagger: bool = False) -> Program:
    """
    A T in terms of RZ(theta)
    """
    if dagger:
        return Program(RZ(-np.pi / 4, q))
    else:
        return Program(RZ(np.pi / 4, q))


def _X(q: QubitDesignator) -> Program:
    """
    An X in terms of RX(pi)

    .. note:
        This introduces a global phase! Don't control this gate.
    """
    p = Program()
    p += RX(np.pi, q)
    return p


def _Y(q: QubitDesignator) -> Program:
    """
    A Y in terms of _RY
    """
    p = Program()
    p += _RY(np.pi, q)
    return p


def _Z(q: QubitDesignator) -> Program:
    """
    A Z in terms of RZ
    """
    p = Program()
    p += RZ(np.pi, q)
    return p


def is_magic_angle(angle: complex) -> bool:
    """
    Checks to see if an angle is 0, +/-pi/2, or +/-pi.
    """
    return bool(
        np.isclose(np.abs(angle), np.pi / 2)
        or np.isclose(np.abs(angle), np.pi)
        or np.isclose(angle, 0.0)
    )


def basic_compile(program: Program) -> Program:
    """
    A rudimentary but predictable compiler.

    No rewiring or optimization is done by this compilation step. There may be
    some gates that are not yet supported. Gates defined in the input program
    are included without change in the output program.

    :param program: A program to be compiled to native quil with simple
        replacements.
    :return: A program with some of the input non-native quil gates replaced
        with basic native quil gate implementations.
    """
    new_prog = Program()
    new_prog.num_shots = program.num_shots
    new_prog.inst(program.defined_gates)

    for inst in program:
        if isinstance(inst, Gate):
            if inst.name == "CCNOT":
                new_prog += _CCNOT(*inst.qubits)
            elif inst.name == "CNOT":
                new_prog += _CNOT(*inst.qubits)
            # NB: we haven't implemented CPHASE00/01/10
            elif inst.name == "CPHASE":
                angle_param = inst.params[0]
                new_prog += _CPHASE(angle_param, *inst.qubits)
            elif inst.name == "CZ":
                new_prog += CZ(*inst.qubits)  # remove dag modifiers
            elif inst.name == "H":
                new_prog += _H(inst.qubits[0])
            elif inst.name == "I":
                new_prog += I(inst.qubits[0])  # remove dag modifiers
            elif inst.name == "ISWAP":
                new_prog += _ISWAP(*inst.qubits)  # remove dag modifiers
            elif inst.name == "PHASE":
                angle_param = inst.params[0]
                new_prog += _PHASE(angle_param, inst.qubits[0])
            elif inst.name == "RX":
                angle_param = cast(complex, inst.params[0])
                if is_magic_angle(angle_param):
                    # in case dagger
                    new_prog += RX(angle_param, inst.qubits[0])
                else:
                    new_prog += _RX(angle_param, inst.qubits[0])
            elif inst.name == "RY":
                angle_param = inst.params[0]
                new_prog += _RY(angle_param, inst.qubits[0])
            elif inst.name == "RZ":
                # in case dagger
                angle_param = inst.params[0]
                new_prog += RZ(angle_param, inst.qubits[0])
            elif inst.name == "S":
                new_prog += _S(inst.qubits[0])
            # NB: we haven't implemented CSWAP or PSWAP
            elif inst.name == "SWAP":
                new_prog += _SWAP(*inst.qubits)
            elif inst.name == "T":
                new_prog += _T(inst.qubits[0])
            elif inst.name == "X":
                new_prog += _X(inst.qubits[0])
            elif inst.name == "XY":
                angle_param = inst.params[0]
                new_prog += XY(angle_param, *inst.qubits)
            elif inst.name == "Y":
                new_prog += _Y(inst.qubits[0])
            elif inst.name == "Z":
                new_prog += _Z(inst.qubits[0])
            elif inst.name in [gate.name for gate in new_prog.defined_gates]:
                new_prog += inst
            else:
                raise ValueError(f"Unknown gate instruction {inst}")
        else:
            new_prog += inst

    new_prog.native_quil_metadata = {  # type: ignore[assignment]
        "final_rewiring": None,
        "gate_depth": None,
        "gate_volume": None,
        "multiqubit_gate_depth": None,
        "program_duration": None,
        "program_fidelity": None,
        "topological_swaps": 0,
    }
    return new_prog



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions for mapping circuits to (near) Clifford circuits."""

from typing import List

import cirq
import numpy as np
import numpy.typing as npt
from cirq.circuits import Circuit

from mitiq.interface import accept_any_qprogram_as_input

# Z gates with these angles/exponents are Clifford gates.
_CLIFFORD_EXPONENTS = np.array([0.0, 0.5, 1.0, 1.5])
_CLIFFORD_ANGLES = [exponent * np.pi for exponent in _CLIFFORD_EXPONENTS]


@accept_any_qprogram_as_input
def is_clifford(circuit: Circuit) -> bool:
    """Returns True if the input argument is Clifford, else False.

    Args:
        circuit: A single operation, list of operations, or circuit.
    """
    return all(
        cirq.has_stabilizer_effect(op) for op in circuit.all_operations()
    )


@accept_any_qprogram_as_input
def count_non_cliffords(circuit: Circuit) -> int:
    """Returns the number of non-Clifford operations in the circuit. Assumes
    the circuit consists of only Rz, Rx, and CNOT operations.

    Args:
        circuit: Circuit to count the number of non-Clifford operations in.
    """
    return sum(
        not cirq.has_stabilizer_effect(op) for op in circuit.all_operations()
    )


def random_clifford(
    num_angles: int, random_state: np.random.RandomState
) -> npt.NDArray[np.float64]:
    """Returns an array of Clifford angles chosen uniformly at random.

    Args:
        num_angles: Number of Clifford angles to return in array.
        random_state: Random state for sampling.
    """
    return np.array(
        [random_state.choice(_CLIFFORD_ANGLES) for _ in range(num_angles)]
    )


@np.vectorize
def closest_clifford(angles: npt.NDArray[np.float64]) -> float:
    """Returns the nearest Clifford angles to the input angles.

    Args:
        non_Clifford_ops: Non-Clifford opperations.
    """
    ang_scaled = angles / (np.pi / 2)
    # if just one min value, return the corresponding nearest cliff.
    if (
        abs((ang_scaled / 0.5) - 1) > 10 ** (-6)
        and abs((ang_scaled / 0.5) - 3) > 10 ** (-6)
        and (abs((ang_scaled / 0.5) - 5) > 10 ** (-6))
    ):
        index = int(np.round(ang_scaled)) % 4
        return _CLIFFORD_ANGLES[index]
    # If equidistant between two Clifford angles, randomly choose one.
    else:
        index_list = [ang_scaled - 0.5, ang_scaled + 0.5]
        index = int(np.random.choice(index_list))
        return _CLIFFORD_ANGLES[index]


@np.vectorize
def is_clifford_angle(
    angles: npt.NDArray[np.float64],
    tol: float = 10**-5,
) -> bool:
    """Function to check if a given angle is Clifford.

    Args:
        angles: rotation angle in the Rz gate.
    """
    angles = angles % (2 * np.pi)
    closest_clifford_angle = closest_clifford(angles)
    if abs(closest_clifford_angle - angles) < tol:
        return True
    return False


@np.vectorize
def angle_to_proximities(angle: float, sigma: float) -> List[float]:
    """Returns probability distribution based on distance from angles to
    Clifford gates.

    Args:
        angle: angle to form probability distribution.

    Returns:
        discrete value of probability distribution calculated from
        exp(-(diff/sigma)^2) where diff is the distance from each angle and the
        Clifford gates.
    """
    s_matrix = cirq.unitary(cirq.S)
    rz_matrix = cirq.unitary(cirq.rz(angle % (2 * np.pi)))
    # TODO: Update loop / if.
    dists = []
    for exponent in range(4):
        if exponent == 0:
            exponent = 4
        diff = np.linalg.norm(rz_matrix - s_matrix**exponent)
        dists.append(np.exp(-((diff / sigma) ** 2)))
    return dists


@np.vectorize
def angle_to_proximity(angle: float, sigma: float) -> float:
    """Returns probability distribution based on distance from angles to
    Clifford gates.

    Args:
        angle: angle to form probability distribution.

    Returns:
        discrete value of probability distribution calculated from
        exp(-(dist/sigma)^2) where dist = sum(dists) is the
        sum of distances from each Clifford gate.
    """
    dists = angle_to_proximities(angle, sigma)
    return np.max(dists)


@np.vectorize
def probabilistic_angle_to_clifford(
    angles: float,
    sigma: float,
    random_state: np.random.RandomState,
) -> npt.NDArray[np.float64]:
    """Returns a Clifford angle sampled from the distribution

                        prob = exp(-(dist/sigma)^2)

    where dist is the Frobenius norm from the 4 clifford angles and the gate
    of interest.

    Args:
        angles: Non-Clifford angles.
        sigma: Width of probability distribution.
    """

    dists = angle_to_proximities(angles, sigma)

    cliff_ang = random_state.choice(
        _CLIFFORD_ANGLES, 1, replace=False, p=np.array(dists) / np.sum(dists)
    )
    return cliff_ang




"""Methods for scaling noise in circuits by layers as required for LRE."""

from mitiq.lre.multivariate_scaling.layerwise_folding import (
    multivariate_layer_scaling,
    get_scale_factor_vectors,
)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""API for using Clifford Data Regression (CDR) error mitigation."""

from functools import wraps
from typing import Any, Callable, List, Optional, Sequence, Union

import numpy as np
from scipy.optimize import curve_fit

from mitiq import QPROGRAM, Executor, Observable, QuantumResult
from mitiq.cdr import (
    generate_training_circuits,
    linear_fit_function,
    linear_fit_function_no_intercept,
)
from mitiq.cdr.clifford_utils import is_clifford
from mitiq.zne.scaling import fold_gates_at_random


def execute_with_cdr(
    circuit: QPROGRAM,
    executor: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    observable: Optional[Observable] = None,
    *,
    simulator: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    num_training_circuits: int = 10,
    fraction_non_clifford: float = 0.1,
    fit_function: Callable[..., float] = linear_fit_function,
    num_fit_parameters: Optional[int] = None,
    scale_factors: Sequence[float] = (1,),
    scale_noise: Callable[[QPROGRAM, float], QPROGRAM] = fold_gates_at_random,
    **kwargs: Any,
) -> float:
    """Function for the calculation of an observable from some circuit of
    interest to be mitigated with CDR (or vnCDR) based on
    Ref. :cite:`Czarnik_2021_Quantum` and Ref. :cite:`Lowe_2021_PRR`.

    The circuit of interest must be compiled in the native basis of the IBM
    quantum computers, that is {Rz, sqrt(X), CNOT}, or such that all the
    non-Clifford gates are contained in the Rz rotations.

    The observable/s to be calculated should be input as an array or a list of
    arrays representing the diagonal of the observables to be measured. Note
    these observables MUST be diagonal in z-basis measurements corresponding to
    the circuit of interest.

    Returns mitigated observables list of raw observables (at noise scale
    factors).

    This function returns the mitigated observable/s.

    Args:
        circuit: Quantum program to execute with error mitigation.
        executor: Executes a circuit and returns a `QuantumResult`.
        observable: Observable to compute the expectation value of.
            If None, the `executor` must return an expectation value. Otherwise
            the `QuantumResult` returned by `executor` is used to compute the
            expectation of the observable.
        simulator: Executes a circuit without noise and returns a
            `QuantumResult`. For CDR to be efficient, the simulator must
            be able to efficiently simulate near-Clifford circuits.
        num_training_circuits: Number of training circuits to be used in the
            mitigation.
        fraction_non_clifford: The fraction of non-Clifford gates to be
            substituted in the training circuits.
        fit_function: The function to map noisy to exact data. Takes array of
            noisy and data and parameters returning a float. See
            ``cdr.linear_fit_function`` for an example.
        num_fit_parameters: The number of parameters the fit_function takes.
        scale_noise: Function for scaling the noise of a quantum circuit.
        scale_factors: Factors by which to scale the noise.

            - When 1.0 is the only scale factor, the method is known as CDR.

            - Note: When scale factors larger than 1.0 are provided, the method
              is known as "variable-noise CDR."

        kwargs: Available keyword arguments are:

            - method_select (string): Specifies the method used to select the
              non-Clifford gates to replace when constructing the near-Clifford
              training circuits. Can be 'uniform' or 'gaussian'.

            - method_replace (string): Specifies the method used to replace the
              selected non-Clifford gates with a Clifford when constructing the
              near-Clifford training circuits. Can be 'uniform', 'gaussian', or
              'closest'.

            - sigma_select (float): Width of the Gaussian distribution used for
              ``method_select='gaussian'``.

            - sigma_replace (float): Width of the Gaussian distribution used
              for ``method_replace='gaussian'``.

            - random_state (int): Seed for sampling.
    """
    # Handle keyword arguments for generating training circuits.

    method_select = kwargs.get("method_select", "uniform")
    method_replace = kwargs.get("method_replace", "closest")
    random_state = kwargs.get("random_state", None)
    kwargs_for_training_set_generation = {
        "sigma_select": kwargs.get("sigma_select"),
        "sigma_replace": kwargs.get("sigma_replace"),
    }

    if num_fit_parameters is None:
        if fit_function is linear_fit_function:
            num_fit_parameters = 1 + len(scale_factors)
        elif fit_function is linear_fit_function_no_intercept:
            num_fit_parameters = len(scale_factors)
        else:
            raise ValueError(
                "Must provide `num_fit_parameters` for custom fit function."
            )

    # cast executor and simulator inputs to Executor type
    if not isinstance(executor, Executor):
        executor = Executor(executor)

    if not isinstance(simulator, Executor):
        simulator = Executor(simulator)

    # Check if circuit is already Clifford
    if is_clifford(circuit):
        return simulator.evaluate(circuit, observable)[0].real

    # Generate training circuits.
    training_circuits = generate_training_circuits(
        circuit,
        num_training_circuits,
        fraction_non_clifford,
        method_select,
        method_replace,
        random_state,
        kwargs=kwargs_for_training_set_generation,
    )

    # [Optionally] Scale noise in circuits.
    all_circuits = [
        [scale_noise(c, s) for s in scale_factors]
        for c in [circuit] + training_circuits  # type: ignore
    ]

    to_run = [circuit for circuits in all_circuits for circuit in circuits]
    all_circuits_shape = (len(all_circuits), len(all_circuits[0]))

    results = executor.evaluate(to_run, observable)
    noisy_results = np.array(results).reshape(all_circuits_shape)

    results = simulator.evaluate(training_circuits, observable)
    ideal_results = np.array(results)

    # Do the regression.
    fitted_params, _ = curve_fit(
        lambda x, *params: fit_function(x, params),
        noisy_results[1:, :].T,
        ideal_results,
        p0=np.zeros(num_fit_parameters),
    )
    return fit_function(noisy_results[0, :], fitted_params)


def mitigate_executor(
    executor: Callable[[QPROGRAM], QuantumResult],
    observable: Optional[Observable] = None,
    *,
    simulator: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    num_training_circuits: int = 10,
    fraction_non_clifford: float = 0.1,
    fit_function: Callable[..., float] = linear_fit_function,
    num_fit_parameters: Optional[int] = None,
    scale_factors: Sequence[float] = (1,),
    scale_noise: Callable[[QPROGRAM, float], QPROGRAM] = fold_gates_at_random,
    **kwargs: Any,
) -> Callable[[QPROGRAM], float]:
    """Returns a clifford data regression (CDR) mitigated version of
    the input 'executor'.

    The input `executor` executes a circuit with an arbitrary backend and
    produces an expectation value (without any error mitigation). The returned
    executor executes the circuit with the same backend but uses clifford
    data regression to produce the CDR estimate of the ideal expectation
    value associated to the input circuit.

    Args:
        executor: Executes a circuit and returns a `QuantumResult`.
        observable: Observable to compute the expectation value of.
            If None, the `executor` must return an expectation value. Otherwise
            the `QuantumResult` returned by `executor` is used to compute the
            expectation of the observable.
        simulator: Executes a circuit without noise and returns a
            `QuantumResult`. For CDR to be efficient, the simulator must
            be able to efficiently simulate near-Clifford circuits.
        num_training_circuits: Number of training circuits to be used in the
            mitigation.
        fraction_non_clifford: The fraction of non-Clifford gates to be
            substituted in the training circuits.
        fit_function: The function to map noisy to exact data. Takes array of
            noisy and data and parameters returning a float. See
            ``cdr.linear_fit_function`` for an example.
        num_fit_parameters: The number of parameters the fit_function takes.
        scale_noise: Function for scaling the noise of a quantum circuit.
        scale_factors: Factors by which to scale the noise.

            - When 1.0 is the only scale factor, the method is known as CDR.

            - Note: When scale factors larger than 1.0 are provided, the method
              is known as "variable-noise CDR."

        kwargs: Available keyword arguments are:

            - method_select (string): Specifies the method used to select the
              non-Clifford gates to replace when constructing the near-Clifford
              training circuits. Can be 'uniform' or 'gaussian'.

            - method_replace (string): Specifies the method used to replace
              the selected non-Clifford gates with a Clifford when constructing
              the near-Clifford training circuits. Can be 'uniform', 'gaussian'
              , or 'closest'.

            - sigma_select (float): Width of the Gaussian distribution used for
              ``method_select='gaussian'``.

            - sigma_replace (float): Width of the Gaussian distribution used
              for ``method_replace='gaussian'``.

            - random_state (int): Seed for sampling."""
    executor_obj = Executor(executor)
    if not executor_obj.can_batch:

        @wraps(executor)
        def new_executor(
            circuit: QPROGRAM,
        ) -> float:
            return execute_with_cdr(
                circuit,
                executor,
                observable,
                simulator=simulator,
                num_training_circuits=num_training_circuits,
                fraction_non_clifford=fraction_non_clifford,
                fit_function=fit_function,
                num_fit_parameters=num_fit_parameters,
                scale_factors=scale_factors,
                scale_noise=scale_noise,
                **kwargs,
            )

    else:

        @wraps(executor)
        def new_executor(
            circuits: List[QPROGRAM],
        ) -> List[float]:
            return [
                execute_with_cdr(
                    circuit,
                    executor,
                    observable,
                    simulator=simulator,
                    num_training_circuits=num_training_circuits,
                    fraction_non_clifford=fraction_non_clifford,
                    fit_function=fit_function,
                    num_fit_parameters=num_fit_parameters,
                    scale_factors=scale_factors,
                    scale_noise=scale_noise,
                    **kwargs,
                )
                for circuit in circuits
            ]

    return new_executor


def cdr_decorator(
    observable: Optional[Observable] = None,
    *,
    simulator: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    num_training_circuits: int = 10,
    fraction_non_clifford: float = 0.1,
    fit_function: Callable[..., float] = linear_fit_function,
    num_fit_parameters: Optional[int] = None,
    scale_factors: Sequence[float] = (1,),
    scale_noise: Callable[[QPROGRAM, float], QPROGRAM] = fold_gates_at_random,
    **kwargs: Any,
) -> Callable[
    [Callable[[Union[QPROGRAM, Any, Any, Any]], QuantumResult]],
    Callable[
        [Union[QPROGRAM, Any, Any, Any]],
        float,
    ],
]:
    """Decorator which adds clifford data regression (CDR) mitigation
    to an executor function, i.e., a function which executes a quantum circuit
    with an arbitrary backend and returns the CDR estimate of the ideal
    expectation value associated to the input circuit.

    Args:
        executor: Executes a circuit and returns a `QuantumResult`.
        observable: Observable to compute the expectation value of.
            If None, the `executor` must return an expectation value. Otherwise
            the `QuantumResult` returned by `executor` is used to compute the
            expectation of the observable.
        simulator: Executes a circuit without noise and returns a
            `QuantumResult`. For CDR to be efficient, the simulator must
            be able to efficiently simulate near-Clifford circuits.
        num_training_circuits: Number of training circuits to be used in the
            mitigation.
        fraction_non_clifford: The fraction of non-Clifford gates to be
            substituted in the training circuits.
        fit_function: The function to map noisy to exact data. Takes array of
            noisy and data and parameters returning a float. See
            ``cdr.linear_fit_function`` for an example.
        num_fit_parameters: The number of parameters the fit_function takes.
        scale_noise: Function for scaling the noise of a quantum circuit.
        scale_factors: Factors by which to scale the noise.

            - When 1.0 is the only scale factor, the method is known as CDR.

            - Note: When scale factors larger than 1.0 are provided, the method
              is known as "variable-noise CDR."

        kwargs: Available keyword arguments are:

            - method_select (string): Specifies the method used to select the
              non-Clifford gates to replace when constructing the near-Clifford
              training circuits. Can be 'uniform' or 'gaussian'.

            - method_replace (string): Specifies the method used to replace the
              selected non-Clifford gates with a Clifford when constructing the
              near-Clifford training circuits. Can be 'uniform', 'gaussian', or
              'closest'.

            - sigma_select (float): Width of the Gaussian distribution used for
              ``method_select='gaussian'``.

            - sigma_replace (float): Width of the Gaussian distribution used
              for ``method_replace='gaussian'``.

            - random_state (int): Seed for sampling.
    """

    def decorator(
        executor: Callable[[QPROGRAM], QuantumResult],
    ) -> Callable[[QPROGRAM], float]:
        return mitigate_executor(
            executor,
            observable,
            simulator=simulator,
            num_training_circuits=num_training_circuits,
            fraction_non_clifford=fraction_non_clifford,
            fit_function=fit_function,
            num_fit_parameters=num_fit_parameters,
            scale_factors=scale_factors,
            scale_noise=scale_noise,
            **kwargs,
        )

    return decorator



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Union, cast

import cirq
import numpy as np
from numpy.typing import NDArray

from mitiq.observable import Observable


def _copy_circuit_parallel(
    circuit: cirq.Circuit, num_copies: int = 2
) -> cirq.Circuit:
    """Given a circuit that acts on N qubits, this function returns a circuit
    that copies the circuit num_copies times in parallel. This means the
    resulting circuit has N * num_copies qubits.

    Args:
        circuit: The circuit to be copied.
        num_copies: The number of copies of circuit to be made.

    Returns:
        A cirq circuit that is the parallel composition of
          num_copies copies of circuit.
    """

    new_circuit = cirq.Circuit()
    N = len(circuit.all_qubits())
    qubits = list(circuit.all_qubits())

    # LineQubits
    if isinstance(qubits[0], cirq.LineQubit):

        def map_for_line_qubits(q: cirq.Qid) -> cirq.Qid:
            assert isinstance(q, cirq.LineQubit)
            return cirq.LineQubit(q.x + N * i)

        for i in range(num_copies):
            new_circuit += circuit.transform_qubits(map_for_line_qubits)

    # GridQubits
    elif isinstance(qubits[0], cirq.GridQubit):
        qubits_cast_grid = cast(List[cirq.GridQubit], qubits)
        grid_rows = max([qu.row + 1 for qu in qubits_cast_grid])

        def map_for_grid_qubits(qu: cirq.Qid) -> cirq.Qid:
            assert isinstance(qu, cirq.GridQubit)
            return cirq.GridQubit(qu.row + grid_rows * i, qu.col)

        for i in range(num_copies):
            new_circuit += circuit.transform_qubits(map_for_grid_qubits)

    return new_circuit


def _apply_diagonalizing_gate(
    circuit: cirq.Circuit, num_copies: int
) -> cirq.Circuit:
    """Apply the diagonalizing gate to a circuit, assuming the circuit has been
    copied in parallel ``num_copies`` times.

    The diagonalizing gate is applied to the first qubit of all copies, to the
    second qubit of all copies, etc.

    Args:
        circuit: The circuit to apply the diagonalizing gate to.
        num_copies: The number of copies of the original circuit that the input
            circuit consists of.

    Returns:
        The circuit with the diagonalizing gate applied.
    """

    new_circuit = circuit.copy()

    # this is the number of qubits in the original circuit
    N = len(circuit.all_qubits()) // num_copies

    diag_gate = _generate_diagonalizing_gate(num_copies)

    for i in range(N):
        qubits = [
            cirq.LineQubit(i + N * j) for j in range(num_copies)
        ]  # select qubit i of each copy

        new_circuit.append(diag_gate(*qubits))

    return new_circuit


def _generate_diagonalizing_gate(num_copies: int = 2) -> cirq.Gate:
    """Generate the diagonalizing gate for the VD algorithm. Currently only
    ``num_copies=2`` is supported.

    Args:
        num_qubits: The number of qubits that the gate acts on.

    Returns: The diagonalizing gate.
    """
    if num_copies == 2:
        diagonalizing_matrix = np.array(
            [
                [1, 0, 0, 0],
                [0, np.sqrt(2) / 2, np.sqrt(2) / 2, 0],
                [0, np.sqrt(2) / 2, -np.sqrt(2) / 2, 0],
                [0, 0, 0, 1],
            ]
        )
    else:
        raise NotImplementedError(
            "Only num_copies = 2 is currently supported."
        )

    return cirq.MatrixGate(diagonalizing_matrix)


def _apply_cyclic_system_permutation(
    matrix: NDArray[np.complex64], N_qubits: int, num_registers: int = 2
) -> NDArray[np.complex64]:
    """
    Function that shifts the rows of a matrix or vector in such a way,
    that each of the num_registers registers of N_qubit qubits are shifted
    cyclically.The implementation is identical to left multiplication
    with repeated swap gates, however this optimisation in considerably
    faster.

    Args:
        matrix: The matrix or vector that should be shifted.
        N_qubits: The number of qubits in each register.
        num_registers: The number of registers.

    Returns:
        The matrix or vector with the rows shifted cyclically.
    """
    matrix = np.array(matrix)

    # determine the row permutation for the cyclic shift operation
    permutation = [
        j + i
        for j in range(2**N_qubits)
        for i in range(0, 2 ** (num_registers * N_qubits), 2**N_qubits)
    ]

    # Some fancy index magic to permute the rows in O(n) time
    # and space (n=2**(num_registers*N_qubits))
    idx = np.empty_like(permutation)
    idx[permutation] = np.arange(len(permutation))

    if matrix.ndim == 2:
        matrix = matrix[idx]
    elif matrix.ndim == 3:
        matrix[:] = matrix[:, idx]
    else:
        raise TypeError("matrix should be 2D or 3D ndarray")
    return matrix


def _apply_symmetric_observable(
    matrix: NDArray[np.complex64],
    N_qubits: int,
    observable: Optional[Union[Observable, NDArray[np.complex64]]] = None,
) -> NDArray[np.complex64]:
    """
    Function that applies a symmetric observable to a matrix or vector.

    Args:
        matrix: The matrix or vector that should be shifted.
        N_qubits: The number of qubits in each register.
        observable: The observable that should be applied.
            If None, the Z observable is used.
        num_registers: The number of registers.

    Returns:
        The matrix or vector with the observable applied.
    """
    z_matrix = np.array([[1.0, 0.0], [0.0, -1.0]])

    if observable is None or (
        isinstance(observable, np.ndarray)
        and np.allclose(observable, z_matrix)
    ):
        # use the default Z observable
        sym_observable_diagonals: List[NDArray[np.complex64]] = []
        for i in range(N_qubits):
            observable_i_diagonal = np.array(
                [
                    j
                    for _ in range(2 ** (i))
                    for j in [1.0, -1.0]
                    for _ in range(2 ** (N_qubits - i - 1))
                ]
            )

            # turn [a, b, c] into [a,a,a,b,b,b,c,c,c]. This is the same as
            # tensoring the N_qubit identity on the right
            observable_i_diagonal_system1 = np.array(
                [observable_i_diagonal for _ in range(2**N_qubits)]
            ).flatten("F")
            # turn [a,b,c] into [a,b,c,a,b,c,a,b,c]. This is the same as
            # tensoring the N_qubit identity on the left
            observable_i_diagonal_system2 = np.array(
                [observable_i_diagonal for _ in range(2**N_qubits)]
            ).flatten("C")
            # add the symmetric observable
            # Create the combined diagonal and explicitly convert to complex64
            combined_diagonal = (
                observable_i_diagonal_system1 + observable_i_diagonal_system2
            ) / 2
            sym_observable_diagonals.append(
                combined_diagonal.astype(np.complex64)
            )

        if matrix.ndim == 2:
            return np.array([sod * matrix for sod in sym_observable_diagonals])
        elif matrix.ndim == 3:
            return np.array(
                [
                    sod * mat
                    for sod in sym_observable_diagonals
                    for mat in matrix
                ]
            )
        else:
            raise ValueError("matrix should be 2D or 3D ndarray")

    else:
        obs_array = (
            observable
            if isinstance(observable, np.ndarray)
            else observable.matrix()
        )
        sym_observable_matrices: List[NDArray[np.complex64]] = []
        for i in range(N_qubits):
            observable_i_matrix = np.kron(
                np.kron(np.eye(2**i), obs_array),
                np.eye(2 ** (N_qubits - i - 1)),
            )
            sym_observable_matrix = (
                np.kron(observable_i_matrix, np.eye(2**N_qubits))
                + np.kron(np.eye(2**N_qubits), observable_i_matrix)
            ) / 2
            sym_observable_matrices.append(
                sym_observable_matrix.astype(np.complex64)
            )

        return np.array(sym_observable_matrices) @ matrix
